/*----------------------------------------------------------------------
  File    : pycoco.c
  Contents: COntinuous time ClOsed Neuron Assembly Detection for Python
  Author  : Christian Borgelt
  History : 2013.02.22 file created
            2013.02.25 trains and database generalized to any iterable
            2013.03.19 adapted to CoCoNAD algorithm variant 'cmrepo'
            2013.03.28 adapted to introduction of types ITEM and SUPP
            2013.05.08 memory leak in tns_fromPyObj() fixed
            2013.12.05 bug in hash code check fixed (-1 vs. negative)
            2013.12.04 pattern spectrum generation function added
            2013.12.07 returned pattern spectrum changed to dictionary
            2013.12.09 multi-threading pattern spectrum creation added
            2013.12.11 signal handling for ctrl-C interrupt added
            2013.12.21 added setting seed to time if it is zero
            2014.01.08 memory leak in py_patspec() fixed (psp)
            2014.03.06 estimation of a pattern spectrum added
            2014.05.09 surrogate data via block permutations added
            2014.05.12 support border added as coconad argument
            2014.05.13 bug in function py_estpsp() fixed (normalization)
            2014.06.12 bug in function py_patspec() fixed (NULL init.)
            2014.07.16 indicator for sequences as target added
            2014.08.28 adapted to modified coconad function interface
            2014.10.11 argument order of patspec() and estpsp() changed
            2014.10.16 functions repinit() and repterm() added
            2014.10.24 changed from LGPL license to MIT license
            2014.12.11 library variant with overlap based support added
            2014.12.16 interval list intersection variant added
            2015.08.07 ensured signal handler removal on error
            2015.08.13 function py_patspec() renamed to py_genpsp()
            2015.08.13 adapted to generic function genpsp()
            2015.08.15 function py_psp2bdr() added (decision border)
            2015.08.16 function py_patred() added (pat. set reduction)
            2015.08.19 adapted to generic pattern set reduction
            2015.08.21 format '|' for pattern spectra added (3 arrays)
            2015.09.03 reporting of associated values made more flexible
            2015.09.04 item cover similarity function cocojim() added
            2016.03.22 perfect extension flag replaced by mode
            2016.03.24 combined integer and graded support added
            2016.10.03 adapted to extensions of coconad implementation
            2016.10.21 adapted to interface changes of coconad impl.
            2016.10.23 closed/maximal repository coded in operation mode
            2016.11.18 coconad miner object and interface introduced
----------------------------------------------------------------------*/
#include <assert.h>
#include <float.h>
#if defined _WIN32 && !defined HAVE_ROUND
#define HAVE_ROUND 1
#endif
#include <Python.h>
#ifndef TRN_SURR
#define TRN_SURR
#endif
#ifndef PSP_ESTIM
#define PSP_ESTIM
#endif
#ifndef PSP_TRAIN
#define PSP_TRAIN
#endif
#ifndef ISR_PATSPEC
#define ISR_PATSPEC
#endif
#ifndef ISR_CLOMAX
#define ISR_CLOMAX
#endif
#include "sigint.h"
#include "random.h"
#include "coconad.h"
#include "ccnpsp.h"
#include "patred.h"

/*----------------------------------------------------------------------
  Preprocessor Definitions
----------------------------------------------------------------------*/
#if PY_MAJOR_VERSION >= 3
#define PyInt_Check     PyLong_Check
#define PyInt_AsLong    PyLong_AsLong
#define PyInt_FromLong  PyLong_FromLong
#else
#define Py_hash_t       long    /* type was introduced with Python 3 */
#endif

/* --- error handling --- */
#define ERR_RTN()       error(0, NULL)
#define ERR_VALUE(s)    error(PyExc_ValueError,   s)
#define ERR_TYPE(s)     error(PyExc_TypeError,    s)
#define ERR_MEM()       error(PyExc_MemoryError,  "not enough memory")
#define ERR_ABORT()     error(PyExc_RuntimeError, "user abort")

/*----------------------------------------------------------------------
  Type Definitions
----------------------------------------------------------------------*/
typedef struct {                /* --- item set report data --- */
  PyObject *res;                /* constructed result */
  int      err;                 /* error flag */
  int      mode;                /* value reporting mode */
  int      cnt;                 /* number of value indicators */
  CCHAR    *rep;                /* indicators of values to report */
} REPDATA;                      /* (item set report data) */

/*----------------------------------------------------------------------
  Item Functions
----------------------------------------------------------------------*/

static size_t hashitem (const void *a, int type)
{                               /* --- compute hash code */
  return (size_t)PyObject_Hash(*(PyObject**)a);
}  /* hashitem() */

/*--------------------------------------------------------------------*/

static int cmpitems (const void *a, const void *b, void *data)
{                               /* --- compare two Python objects */
  return PyObject_RichCompareBool(*(PyObject**)a,*(PyObject**)b,Py_NE);
}  /* cmpitems() */             /* return 0 if objects are equal */

/*----------------------------------------------------------------------
  Parameter Functions
----------------------------------------------------------------------*/

static int get_target (const char *s)
{                               /* --- get target */
  int t;                        /* sequence flag */

  assert(s);                    /* check the function argument */
  if (*s && !s[1] && !s[2]) {   /* evaluate textual identifier */
    if      (strcmp(s, "set")          == 0) s = "fi";
    else if (strcmp(s, "sets")         == 0) s = "fi";
    else if (strcmp(s, "itemset")      == 0) s = "fi";
    else if (strcmp(s, "itemsets")     == 0) s = "fi";
    else if (strcmp(s, "all")          == 0) s = "fi";
    else if (strcmp(s, "allsets")      == 0) s = "fi";
    else if (strcmp(s, "frq")          == 0) s = "fi";
    else if (strcmp(s, "freq")         == 0) s = "fi";
    else if (strcmp(s, "frequent")     == 0) s = "fi";
    else if (strcmp(s, "frqsets")      == 0) s = "fi";
    else if (strcmp(s, "freqsets")     == 0) s = "fi";
    else if (strcmp(s, "cls")          == 0) s = "ci";
    else if (strcmp(s, "clsd")         == 0) s = "ci";
    else if (strcmp(s, "closed")       == 0) s = "ci";
    else if (strcmp(s, "max")          == 0) s = "mi";
    else if (strcmp(s, "maxi")         == 0) s = "mi";
    else if (strcmp(s, "maximal")      == 0) s = "mi";
    else if (strcmp(s, "perm")         == 0) s = "fp";
    else if (strcmp(s, "perms")        == 0) s = "fp";
    else if (strcmp(s, "permutation")  == 0) s = "fp";
    else if (strcmp(s, "permutations") == 0) s = "fp";
    else if (strcmp(s, "allperm")      == 0) s = "fp";
    else if (strcmp(s, "allperms")     == 0) s = "fp";
    else if (strcmp(s, "frqperm")      == 0) s = "fp";
    else if (strcmp(s, "frqperms")     == 0) s = "fp";
    else if (strcmp(s, "freqperm")     == 0) s = "fp";
    else if (strcmp(s, "freqperms")    == 0) s = "fp";
    else if (strcmp(s, "clsperm")      == 0) s = "cp";
    else if (strcmp(s, "clsperms")     == 0) s = "cp";
    else if (strcmp(s, "clsdperm")     == 0) s = "cp";
    else if (strcmp(s, "clsdperms")    == 0) s = "cp";
    else if (strcmp(s, "closedperm")   == 0) s = "cp";
    else if (strcmp(s, "closedperms")  == 0) s = "cp";
    else if (strcmp(s, "maxperm")      == 0) s = "mp";
    else if (strcmp(s, "maxperms")     == 0) s = "mp";
    else if (strcmp(s, "maxiperm")     == 0) s = "mp";
    else if (strcmp(s, "maxiperms")    == 0) s = "mp";
    else if (strcmp(s, "maximalperm")  == 0) s = "mp";
    else if (strcmp(s, "maximalperms") == 0) s = "mp";
    else if (strcmp(s, "seq")          == 0) s = "fs";
    else if (strcmp(s, "seqs")         == 0) s = "fs";
    else if (strcmp(s, "sequence")     == 0) s = "fs";
    else if (strcmp(s, "sequences")    == 0) s = "fs";
    else if (strcmp(s, "allseq")       == 0) s = "fs";
    else if (strcmp(s, "allseqs")      == 0) s = "fs";
    else if (strcmp(s, "frqseq")       == 0) s = "fs";
    else if (strcmp(s, "frqseqs")      == 0) s = "fs";
    else if (strcmp(s, "freqseq")      == 0) s = "fs";
    else if (strcmp(s, "freqseqs")     == 0) s = "fs";
    else if (strcmp(s, "clsseq")       == 0) s = "cs";
    else if (strcmp(s, "clsseqs")      == 0) s = "cs";
    else if (strcmp(s, "clsdseq")      == 0) s = "cs";
    else if (strcmp(s, "clsdseqs")     == 0) s = "cs";
    else if (strcmp(s, "closedseq")    == 0) s = "cs";
    else if (strcmp(s, "closedseqs")   == 0) s = "cs";
    else if (strcmp(s, "maxseq")       == 0) s = "ms";
    else if (strcmp(s, "maxseqs")      == 0) s = "ms";
    else if (strcmp(s, "maxiseq")      == 0) s = "ms";
    else if (strcmp(s, "maxiseqs")     == 0) s = "ms";
    else if (strcmp(s, "maximalseq")   == 0) s = "ms";
    else if (strcmp(s, "maximalseqs")  == 0) s = "ms";
  }
  if (*s && (!s[1] || !s[2])) { /* check for a valid string */
    switch (s[1]) {             /* evaluate the sequence indicator */
      case  0 : t = CCN_ITEMSET;  break;
      case 'i': t = CCN_ITEMSET;  break;
      case 'p': t = CCN_PERMUT;   break;
      case 'o': t = CCN_PERMUT;   break;
      case 's': t = CCN_SEQUENCE; break;
      case 'q': t = CCN_SEQUENCE; break;
      default : t = -1;           break;
    }
    switch (s[0]) {             /* evaluate the target code */
      case 'a': return t|ISR_FREQUENT;
      case 's': return t|ISR_FREQUENT;
      case 'f': return t|ISR_FREQUENT;
      case 'c': return t|ISR_CLOSED;
      case 'm': return t|ISR_MAXIMAL;
    }
  }
  PyErr_SetString(PyExc_ValueError, "invalid target type");
  return -1;                    /* return an error code */
}  /* get_target() */

/*--------------------------------------------------------------------*/
#ifdef CCN_OVERLAP

static int get_sim (const char *s)
{                               /* --- get similarity measure */
  assert(s);                    /* check the function argument */
  if (*s && !s[1] && !s[2]) {   /* evaluate textual identifier */
    if      (strcmp(s, "none")            == 0) s = "x";
    else if (strcmp(s, "russel_rao")      == 0) s = "r";
    else if (strcmp(s, "kulczynski")      == 0) s = "k";
    else if (strcmp(s, "jaccard")         == 0) s = "j";
    else if (strcmp(s, "tanimoto")        == 0) s = "j";
    else if (strcmp(s, "dice")            == 0) s = "d";
    else if (strcmp(s, "sorensen")        == 0) s = "d";
    else if (strcmp(s, "czekanowski")     == 0) s = "d";
    else if (strcmp(s, "sokal_sneath_1")  == 0) s = "s";
    else if (strcmp(s, "sokal_michener")  == 0) s = "m";
    else if (strcmp(s, "hamming")         == 0) s = "m";
    else if (strcmp(s, "faith")           == 0) s = "f";
    else if (strcmp(s, "rogers_tanimoto") == 0) s = "t";
    else if (strcmp(s, "sokal_sneath_2")  == 0) s = "n";
    else if (strcmp(s, "gower_legendre")  == 0) s = "n";
    else if (strcmp(s, "sokal_sneath_3")  == 0) s = "o";
    else if (strcmp(s, "baroni_buser")    == 0) s = "b";
    else if (strcmp(s, "generic")         == 0) s = "g";
  }
  if (*s && !s[1]) {            /* check for a valid string */
    switch (s[0]) {             /* evaluate the target code */
      case 'x': return CCN_NONE;
      case 'r': return CCN_RUSSEL_RAO;
      case 'k': return CCN_KULCZYNSKI;
      case 'j': return CCN_JACCARD;
      case 'd': return CCN_DICE;
      case 's': return CCN_SOKAL_SNEATH_1;
      case 'm': return CCN_SOKAL_MICHENER;
      case 'f': return CCN_FAITH;
      case 't': return CCN_ROGERS_TANIMOTO;
      case 'n': return CCN_SOKAL_SNEATH_2;
      case 'o': return CCN_SOKAL_SNEATH_3;
      case 'b': return CCN_BARONI_BUSER;
      case 'g': return CCN_GENERIC;
    }
  }
  PyErr_SetString(PyExc_ValueError, "invalid similarity measure");
  return -1;                    /* return an error code */
}  /* get_sim() */

#endif
/*--------------------------------------------------------------------*/

static int get_coconad (const char *s, int target)
{                               /* --- get CoCoNAD algorithm */
  assert(s);                    /* check the function argument */
  #ifndef CCN_OVERLAP           /* if standard (MIS) support */
  if (!*s) s = (target & (CCN_PERMUT|CCN_SEQUENCE)) ? "f" : "r";
  #else                         /* if influence map overlap support */
  if (!*s) s = "i";             /* get default algorithm */
  #endif                        /* (differs accord. to support type) */
  if (s[0] && s[1]) {           /* evaluate textual identifier */
    if      (strcmp(s, "basic")      == 0) s = "b";
    else if (strcmp(s, "base")       == 0) s = "b";
    else if (strcmp(s, "filt")       == 0) s = "f";
    else if (strcmp(s, "filter")     == 0) s = "f";
    else if (strcmp(s, "filtered")   == 0) s = "f";
    else if (strcmp(s, "refined")    == 0) s = "f";
    else if (strcmp(s, "repo")       == 0) s = "r";
    else if (strcmp(s, "repository") == 0) s = "r";
    else if (strcmp(s, "cmrepo")     == 0) s = "r";
    #ifdef CCN_OVERLAP
    else if (strcmp(s, "isect")      == 0) s = "i";
    else if (strcmp(s, "intersect")  == 0) s = "i";
    #endif
  }
  if (s[0] && !s[1]) {          /* translate the algorithm string */
    switch (s[0]) {             /* evaluate the algorithm code */
      case 'b': return CCN_BASIC;
      case 'f': return CCN_FILTER;
      case 'r': return CCN_REPO;
      #ifdef CCN_OVERLAP
      case 'i': return CCN_ISECT;
      #endif
    }
  }
  PyErr_SetString(PyExc_ValueError, "invalid CoCoNAD algorithm");
  return -1;                    /* return an error code */
}  /* get_coconad() */

/*--------------------------------------------------------------------*/

static int get_surr (const char *s)
{                               /* --- get surrogate function code */
  assert(s);                    /* check the function argument */
  if (s[0] && s[1]) {           /* evaluate textual identifier */
    if      (strcmp(s, "ident")     == 0) s = "i";
    else if (strcmp(s, "identity")  == 0) s = "i";
    else if (strcmp(s, "random")    == 0) s = "r";
    else if (strcmp(s, "randomize") == 0) s = "r";
    else if (strcmp(s, "dither")    == 0) s = "d";
    else if (strcmp(s, "displace")  == 0) s = "d";
    else if (strcmp(s, "jitter")    == 0) s = "d";
    else if (strcmp(s, "shift")     == 0) s = "s";
    else if (strcmp(s, "kernel")    == 0) s = "k";
    else if (strcmp(s, "kernest")   == 0) s = "k";
    else if (strcmp(s, "perm")      == 0) s = "p";
    else if (strcmp(s, "permute")   == 0) s = "p";
    else if (strcmp(s, "block")     == 0) s = "b";
    else if (strcmp(s, "blocked")   == 0) s = "b";
    else if (strcmp(s, "blkperm")   == 0) s = "b";
    else if (strcmp(s, "blockperm") == 0) s = "b";
  }
  if (s[0] && !s[1]) {          /* translate surrogate method string */
    switch (s[0]) {             /* evaluate the surrogate method code */
      case 'i': return CCN_IDENTITY;
      case 'r': return CCN_RANDOM;
      case 'd': return CCN_DITHER;
      case 's': return CCN_SHIFT;
      case 'k': return CCN_KERNEST;
      case 'p': return CCN_PERMUTE;
      case 'b': return CCN_BLOCK;
    }
  }
  PyErr_SetString(PyExc_ValueError,
                  "invalid surrogate generation method");
  return -1;                    /* return an error code */
}  /* get_surr() */

/*--------------------------------------------------------------------*/

static int get_rand (const char *s)
{                               /* --- get random function code */
  assert(s);                    /* check the function argument */
  if (s[0] && s[1]) {           /* evaluate textual identifier */
    if      (strcmp(s, "uni")         == 0) s = "u";
    else if (strcmp(s, "uniform")     == 0) s = "u";
    else if (strcmp(s, "rect")        == 0) s = "r";
    else if (strcmp(s, "rectangle")   == 0) s = "r";
    else if (strcmp(s, "rectangular") == 0) s = "r";
    else if (strcmp(s, "triang")      == 0) s = "t";
    else if (strcmp(s, "triangle")    == 0) s = "t";
    else if (strcmp(s, "triangular")  == 0) s = "t";
    else if (strcmp(s, "Gauss")       == 0) s = "g";
    else if (strcmp(s, "Gaussian")    == 0) s = "g";
    else if (strcmp(s, "gauss")       == 0) s = "g";
    else if (strcmp(s, "gaussian")    == 0) s = "g";
    else if (strcmp(s, "normal")      == 0) s = "n";
  }
  if (s[0] && !s[1]) {          /* translate surrogate method string */
    switch (s[0]) {             /* evaluate the surrogate method code */
      case 'u': return RNG_UNIFORM;
      case 'r': return RNG_RECT;
      case 't': return RNG_TRIANG;
      case 'g': return RNG_GAUSS;
      case 'n': return RNG_NORMAL;
    }
  }
  PyErr_SetString(PyExc_ValueError,
                  "invalid probability density function");
  return -1;                    /* return an error code */
}  /* get_rand() */

/*--------------------------------------------------------------------*/

static int get_red (const char *s)
{                               /* --- get random function code */
  assert(s);                    /* check the function argument */
  if (s[0] && s[1]) {           /* evaluate textual identifier */
    if      (strcmp(s, "none")        == 0) s = "x";
    else if (strcmp(s, "coins")       == 0) s = "c";
    else if (strcmp(s, "coins0")      == 0) s = "c";
    else if (strcmp(s, "coins1")      == 0) s = "C";
    else if (strcmp(s, "coins+1")     == 0) s = "C";
    else if (strcmp(s, "items")       == 0) s = "i";
    else if (strcmp(s, "items2")      == 0) s = "i";
    else if (strcmp(s, "neurons")     == 0) s = "i";
    else if (strcmp(s, "cover")       == 0) s = "s";
    else if (strcmp(s, "cover0")      == 0) s = "s";
    else if (strcmp(s, "covered")     == 0) s = "s";
    else if (strcmp(s, "covered0")    == 0) s = "s";
    else if (strcmp(s, "cover1")      == 0) s = "S";
    else if (strcmp(s, "covered1")    == 0) s = "S";
    else if (strcmp(s, "leni")        == 0) s = "l";
    else if (strcmp(s, "leni0")       == 0) s = "l";
    else if (strcmp(s, "lenient")     == 0) s = "l";
    else if (strcmp(s, "lenient0")    == 0) s = "l";
    else if (strcmp(s, "leni1")       == 0) s = "L";
    else if (strcmp(s, "lenient1")    == 0) s = "L";
    else if (strcmp(s, "strict")      == 0) s = "t";
    else if (strcmp(s, "strict0")     == 0) s = "t";
    else if (strcmp(s, "strict1")     == 0) s = "T";
  }
  if (s[0] && !s[1]) {          /* translate surrogate method string */
    switch (s[0]) {             /* evaluate the surrogate method code */
      case 'x': return PSR_NONE;
      case 'c': return PSR_COINS0;
      case 'C': return PSR_COINS1;
      case 'i': return PSR_ITEMS2;
      case 's': return PSR_COVER0;
      case 'S': return PSR_COVER1;
      case 'l': return PSR_LENIENT0;
      case 'L': return PSR_LENIENT1;
      case 't': return PSR_STRICT0;
      case 'T': return PSR_STRICT1;
    }
  }
  PyErr_SetString(PyExc_ValueError,
                  "invalid pattern set reduction method");
  return -1;                    /* return an error code */
}  /* get_red() */

/*----------------------------------------------------------------------
  Auxiliary Functions
----------------------------------------------------------------------*/

static void clean1 (PyObject *a)
{ if (a) Py_DECREF(a); }

/*--------------------------------------------------------------------*/

static void clean2 (PyObject *a, PyObject *b)
{ if (a) { Py_DECREF(a); } if (b) { Py_DECREF(b); } }

/*--------------------------------------------------------------------*/

static void clean3 (PyObject *a, PyObject *b, PyObject *c)
{ if (a) { Py_DECREF(a); } if (b) { Py_DECREF(b); }
  if (c) { Py_DECREF(c); } }

/*--------------------------------------------------------------------*/

static void clean4 (PyObject *a, PyObject *b, PyObject *c, PyObject *d)
{ if (a) { Py_DECREF(a); } if (b) { Py_DECREF(b); }
  if (c) { Py_DECREF(c); } if (d) { Py_DECREF(d); } }

/*--------------------------------------------------------------------*/

static void cleanup (TRAINSET *trnset, TRAIN *trn)
{                              /* --- clean up after an error */
  if (trn)    trn_delete(trn); /* drop references and delete */
  if (trnset) tns_delete(trnset, 1);  /* train and train set */
}  /* cleanup() */

/*--------------------------------------------------------------------*/

static void* error (PyObject *type, const char *msg)
{                               /* --- report an error */
  sig_remove();                 /* remove the signal handler */
  if (msg) PyErr_SetString(type, msg); /* set the error message */
  return NULL;                  /* return NULL as an error indicator */
}  /* error() */

/*--------------------------------------------------------------------*/
#ifdef CCN_OVERLAP              /* if influence map overlap support */

static void* sim_fromPyObj (double *pars, PyObject *pypars)
{                               /* --- get similarity parameters */
  int      n;                   /* similarity params. sequence length */
  PyObject *o;                  /* to traverse the parameters */

  assert(pars);                 /* check the function arguments */
  if (!pypars) return (void*)1; /* check for a Python object */
  if (!PySequence_Check(pypars))/* check for a sequence */
    return ERR_TYPE("similarity parameters must be "
                    "a sequence of numbers");
  n = (int)PySequence_Length(pypars);
  if (n > 8) n = 8;             /* get the sequence length */
  while (--n >= 0) {            /* traverse the sequence elements */
    o = PySequence_GetItem(pypars, n);
    if      (PyLong_Check(o))   /* if element is a long integer */
      pars[n] = (double)PyLong_AsLong(o);
    else if (PyInt_Check(o))    /* if element is an integer */
      pars[n] = (double)PyInt_AsLong(o);
    else if (PyFloat_Check(o))  /* if element is a float */
      pars[n] = PyFloat_AsDouble(o);
    else { clean1(o);           /* get the element value as parameter */
      return ERR_TYPE("similarity parameters must be numbers"); }
    Py_DECREF(o);               /* drop the element reference */
  }
  return (void*)1;              /* return 'ok' */
}  /* sim_fromPyObj() */

#endif  /* #ifdef CCN_OVERLAP */
/*--------------------------------------------------------------------*/

static TRAINSET* tns_fromPyObj (PyObject *trains)
{                               /* --- create a train set */
  ITEM      i;                  /* item identifier */
  TID       k, n;               /* loop variables */
  Py_hash_t h;                  /* hash value of item */
  ITEMBASE  *ibase;             /* underlying item base */
  TRAINSET  *trnset;            /* created train set */
  TRAIN     *trn;               /* to create the trains */
  PyObject  *ti, *ei;           /* train and event/spike iterator */
  PyObject  *pair;              /* to traverse (id,train) pairs */
  PyObject  *item;              /* to traverse the items */
  PyObject  *list;              /* to access event/spike time lists */
  PyObject  *s;                 /* to traverse the events/spikes */
  int       dict;               /* whether train set is a dictionary */

  assert(trains);               /* check the function argument */
  ibase = ib_create(IB_OBJNAMES, 0, hashitem, cmpitems, NULL, NULL);
  if (!ibase) return ERR_MEM(); /* create an item base */
  trnset = tns_create(ibase);   /* and a train set */
  if (!trnset) { ib_delete(ibase); return ERR_MEM(); }
  dict = PyDict_Check(trains);  /* check whether train set is dict. */
  ti = PyObject_GetIter(trains);/* get an iterator for the trains */
  if (!ti) { cleanup(trnset, NULL);
    return ERR_TYPE("train database must be iterable"); }
  while ((pair = PyIter_Next(ti)) != NULL) {
    if (dict) {                 /* if dictionary, key is item */
      item = pair;              /* and value is event time list */
      list = PyDict_GetItem(trains, item); Py_INCREF(list); }
    else {                      /* if some other type of iterable */
      if (!PySequence_Check(pair)  /* trains must be pairs/sequences */
      ||  (PySequence_Length(pair) < 2)) {
        clean2(pair, ti); cleanup(trnset, NULL);
        return ERR_TYPE("trains in non-dictionary must be pairs "
                        "(sequences with length >= 2)"); }
      item = PySequence_GetItem(pair, 0);
      if (!item) { clean2(pair,ti); cleanup(trnset,NULL); return NULL; }
      list = PySequence_GetItem(pair, 1);
      Py_DECREF(pair);          /* get the item and the train */
      if (!list) { clean2(item,ti); cleanup(trnset,NULL); return NULL; }
    }
    h = PyObject_Hash(item);    /* check whether item is hashable */
    if (h == -1) { clean3(list, item, ti); cleanup(trnset, NULL);
      return ERR_TYPE("items must be hashable"); }
    i = ib_add(ibase, &item);   /* get the item of the train */
    if (i <   0) { clean3(list, item, ti); cleanup(trnset, NULL);
      return ERR_MEM(); }       /* check whether item could be added */
    Py_DECREF(item);            /* drop the item reference */
    if      (PySequence_Check(list))  /* if sequence, */
      n = (TID)PySequence_Size(list); /* get its length */
    else if (PyAnySet_Check(list))    /* if set or frozenset, */
      n = (TID)PySet_Size(list);      /* get its size */
    else {                      /* if any other iterable object */
      ei = PyObject_GetIter(list);
      if (!ei) { clean2(list, ti); cleanup(trnset, NULL);
        return ERR_TYPE("event/spike time list must be iterable"); }
      for (n = 0; ((s = PyIter_Next(ei)) != NULL); n++)
        Py_DECREF(s);           /* count the number of elements */
      if (PyErr_Occurred()) {   /* (by traversing the iterable) */
        clean3(ei, list, ti); cleanup(trnset, NULL); return NULL; }
    }                           /* check for a Python error */
    trn = trn_create(i,NULL,n); /* create a train object */
    if (!trn) { clean2(list, ti); cleanup(trnset, NULL);
      return ERR_MEM(); }       /* check for proper creation */
    ei = PyObject_GetIter(list);/* create an event time iterator */
    if (!ei)  { clean2(list, ti); cleanup(trnset, NULL);
      return ERR_TYPE("event/spike time list must be iterable"); }
    Py_DECREF(list);            /* drop the list reference */
    for (k = 0; (k < n) && ((s = PyIter_Next(ei)) != NULL); k++) {
      if      (PyLong_Check(s)) /* get integer event time */
        trn->pts[k] = (double)PyLong_AsLong(s);
      else if (PyInt_Check(s))  /* get long event time */
        trn->pts[k] = (double)PyInt_AsLong(s);
      else if (PyFloat_Check(s))/* get floating-point event time */
        trn->pts[k] = PyFloat_AsDouble(s);
      else { clean3(s, ei, ti); cleanup(trnset, trn);
        return ERR_TYPE("event/spike times must be numbers"); }
      Py_DECREF(s);             /* get the event/spike time and */
    }                           /* drop the spike reference */
    trn->len = n;               /* store the length of the train */
    Py_DECREF(ei);              /* drop the event time iterator */
    if (PyErr_Occurred() || (tns_add(trnset, trn) < 0)) {
      clean1(ti); cleanup(trnset, trn); return ERR_MEM(); }
  }
  Py_DECREF(ti);                /* drop the train iterator */
  if (PyErr_Occurred()) { tns_delete(trnset, 1); return NULL; }
  return trnset;                /* return the created train set */
}  /* tns_fromPyObj() */

/*--------------------------------------------------------------------*/

static void* isr_pyborder (ISREPORT *rep, PyObject *border)
{                               /* --- set reporter filtering border */
  Py_ssize_t n;                 /* loop variable, sequence length */
  PyObject   *o;                /* to traverse the border elements */
  double     s;                 /* support threshold as a double */
  RSUPP      supp;              /* support threshold (scaled) */

  assert(rep);                  /* check the function arguments */
  if (!border) return (void*)1; /* check for a border */
  if (!PySequence_Check(border)) {
    return ERR_TYPE("border must be a sequence of numbers"); }
  n = PySequence_Length(border);/* check for a sequence */
  if (n <= 0)  return (void*)1; /* empty sequences need no processing */
  while (n > 0) { --n;          /* traverse the sequence elements */
    o = PySequence_GetItem(border, n);
    if      (PyLong_Check(o))   /* if element is a long integer */
      supp = (RSUPP)PyLong_AsLong(o);
    else if (PyInt_Check(o))    /* if element is an integer */
      supp = (RSUPP)PyInt_AsLong(o);
    else if (PyFloat_Check(o)){ /* if element is a float */
      s = PyFloat_AsDouble(o);
      supp = (s >= (double)RSUPP_MAX) ? RSUPP_MAX : (RSUPP)s; }
    else { clean1(o);           /* get the element value as support */
      return ERR_TYPE("border elements must be numbers"); }
    Py_DECREF(o);               /* drop the element reference */
    if (isr_setbdr(rep, (ITEM)n, supp) < 0) return ERR_MEM();
  }                             /* set the border support */
  return (void*)1;              /* return 'ok' */
}  /* isr_pyborder() */

/*--------------------------------------------------------------------*/

static void isr_toPyObj (ISREPORT *rep, void *data)
{                               /* --- report an item set */
  int      v;                   /* loop variable for values */
  ITEM     i, n;                /* loop variables */
  SUPP     s;                   /* support value for reporting */
  double   e, x;                /* evaluation and scaling factor */
  PyObject *pat;                /* pattern of item set and values */
  PyObject *iset;               /* found item set */
  PyObject *vals;               /* values associated to item set */
  PyObject *obj;                /* to traverse the items, values */
  REPDATA  *rd = data;          /* report data structure */

  assert(rep && data);          /* check the function arguments */
  n    = isr_cnt(rep);          /* get the size of the item set */
  iset = PyTuple_New(n);        /* create an item set tuple */
  if (!iset) { rd->err = -1; return; }
  for (i = 0; i < n; i++) {     /* traverse the items */
    obj = (PyObject*)isr_itemobj(rep, isr_itemx(rep, i));
    Py_INCREF(obj);             /* get the corresp. Python object */
    PyTuple_SET_ITEM(iset, i, obj);
  }                             /* store the item in the set */
  if      (rd->mode == '[') vals = PyList_New (rd->cnt);
  else if (rd->mode == '(') vals = PyTuple_New(rd->cnt);
  else                      vals = PyTuple_New(rd->cnt+1);
  if (!vals) { clean1(iset); rd->err = -1; return; }
  s = 0; e = 0;                 /* initialize the report variables */
  for (v = 0; v < rd->cnt; v++){/* traverse the values to store */
    switch (rd->rep[v]) {       /* evaluate the value indicator */
      #ifdef CCN_OVERLAP        /* if floating point support */
      case 'a': e = (double)isr_supp(rep); x =   1.0; break;
      case 's': e = (double)isr_supp(rep); x =   1.0; break;
      case 'S': e = (double)isr_supp(rep); x = 100.0; break;
      case 'w': e = (double)isr_wgt(rep);  x =   1.0; break;
      case 'W': e = (double)isr_wgt(rep);  x = 100.0; break;
      #else                     /* if integer support */
      case 'a': s = (SUPP)  isr_supp(rep); x =   0.0; break;
      case 's': s = (SUPP)  isr_supp(rep); x =   0.0; break;
      #endif
      default : s = 0;                     x =   0.0; break;
    }                           /* get the requested value */
    if (x == 0) obj = PyInt_FromLong((long)s);
    else        obj = PyFloat_FromDouble(e *x);
    if (!obj) { clean2(iset, vals); rd->err = -1; return; }
    if      (rd->mode == '[') PyList_SET_ITEM (vals, v,   obj);
    else if (rd->mode == '(') PyTuple_SET_ITEM(vals, v,   obj);
    else                      PyTuple_SET_ITEM(vals, v+1, obj);
  }                             /* store the created value */
  if (rd->mode == 0)            /* if to store in pattern object */
    pat = vals;                 /* just get the value tuple */
  else {                        /* if to store in separate object */
    pat = PyTuple_New(2);       /* create a pair of set and values */
    if (!pat) { clean2(iset, vals); rd->err = -1; return; }
    PyTuple_SET_ITEM(pat, 1, vals);
  }                             /* set the pair elements */
  PyTuple_SET_ITEM(pat, 0, iset);
  if (PyList_Append(rd->res, pat) != 0)
    rd->err = -1;               /* append the pair to the result list */
  Py_DECREF(pat);               /* remove internal reference to pair */
}  /* isr_toPyObj() */

/*--------------------------------------------------------------------*/
#ifndef CCN_OVERLAP

static PyObject* psp_toPyObj (PATSPEC *psp, double scale, int format)
{                               /* --- report pattern spectrum */
  int        e = 0;             /* error indicator */
  Py_ssize_t i;                 /* list size, result list index */
  ITEM       size;              /* loop variable for sizes */
  RSUPP      supp, min, max;    /* loop variable for supports */
  size_t     frq;               /* frequency of a pattern signature */
  PyObject   *res;              /* created Python list object */
  PyObject   *cols[3] = {NULL,NULL,NULL};     /* result columns */
  PyObject   *z, *s, *f, *t;    /* (elements of) result triplet */

  assert(psp);                  /* check the function arguments */
  i = (Py_ssize_t)psp_sigcnt(psp);
  if       (format == '#')      /* evaluate pattern spectrum format */
    res = PyDict_New();         /* dictionary (size,supp) -> freq */
  else if ((format == '=') || (format == '-'))
    res = PyList_New(i);        /* list of triplets (size,supp,freq) */
  else {                        /* three columns for size/supp/freq */
    res     = PyList_New(3); if (!res) return NULL;
    cols[0] = PyList_New(i); if (!cols[0]) { clean1(res); return NULL; }
    PyList_SET_ITEM(res, 0, cols[0]);  /* sizes */
    cols[1] = PyList_New(i); if (!cols[1]) { clean1(res); return NULL; }
    PyList_SET_ITEM(res, 1, cols[1]);  /* support values */
    cols[2] = PyList_New(i); if (!cols[2]) { clean1(res); return NULL; }
    PyList_SET_ITEM(res, 2, cols[2]);  /* frequencies */
  }                             /* (create the return object) */
  if (!res) return NULL;        /* check for successful creation */
  for (i = 0, size = psp_min(psp); size <= psp_max(psp); size++) {
    min = psp_min4sz(psp,size); /* traverse the pattern sizes */
    max = psp_max4sz(psp,size); /* get range of support values */
    if (max < min) continue;    /* skip size with empty support range */
    z = PyInt_FromLong((long)size);
    if (!z) { e = -1; break; }  /* create size object for storing */
    for (supp = min; supp <= max; supp++) {
      if ((frq = psp_getfrq(psp, size, supp)) <= 0)
        continue;               /* traverse the support values */
      s = PyInt_FromLong((long)supp);
      if (!s) {                 e = -1; break; }
      f = PyFloat_FromDouble((double)frq *scale);
      if (!f) {   clean1(s);    e = -1; break; }
      if      (format == '#') { /* if result is dictionary */
        t = PyTuple_New(2);     /* create a tuple as a key */
        if (!t) { clean2(f, s); e = -1; break; }
        Py_INCREF(z);           /* get a new size reference */
        PyTuple_SET_ITEM(t, 0, z);  /* create key (size, support) */
        PyTuple_SET_ITEM(t, 1, s);  /* and map key to frequency */
        PyDict_SetItem(res, t, f); }
      else if (format == '=') { /* if result is list of triplets */
        t = PyTuple_New(3);     /* create a result list element */
        if (!t) { clean2(f, s); e = -1; break; }
        Py_INCREF(z);           /* get a new size reference */
        PyTuple_SET_ITEM(t, 0, z);  /* fill element with triplet */
        PyTuple_SET_ITEM(t, 1, s);  /* (size, support, freq) and */
        PyTuple_SET_ITEM(t, 2, f);  /* set element in the result */
        PyList_SET_ITEM(res, i, t); }
      else {                    /* if result is three columns */
        Py_INCREF(z);           /* get a new size reference */
        PyList_SET_ITEM(cols[0], i, z);
        PyList_SET_ITEM(cols[1], i, s);
        PyList_SET_ITEM(cols[2], i, f);
      }                         /* set size, support, freq. in lists */
      i += 1;                   /* count the processed signature */
    }
    Py_DECREF(z);               /* initial reference no longer needed */
    if (e) break;               /* if an error occurred, abort loop */
  }
  if (e) { clean1(res); return NULL; }
  return res;                   /* return created pattern spectrum */
}  /* psp_toPyObj() */

/*--------------------------------------------------------------------*/
#else                           /* #ifndef CCN_OVERLAP .. */

static PyObject* psp_toPyObj (PATSPEC *psp, double scale, int format)
{                               /* --- report pattern spectrum */
  int        e = 0;             /* error indicator */
  Py_ssize_t i;                 /* list size, result list index */
  ITEM       size;              /* loop variable for sizes */
  RSUPP      min, max;          /* minimum and maximum support */
  PyObject   *res;              /* created Python list object */
  PyObject   *cols[3] = {NULL,NULL,NULL};     /* result columns */
  PyObject   *z, *s, *m, *t;    /* (elements of) result triplet */

  assert(psp);                  /* check the function arguments */
  for (i = 0, size = psp_min(psp); size <= psp_max(psp); size++)
    if (psp_max4sz(psp, size) >= psp_min4sz(psp, size))
      i++;                      /* count the used sizes */
  if       (format == '#')      /* evaluate pattern spectrum format */
    res = PyDict_New();         /* dictionary (size,supp) -> freq */
  else if ((format == '=') || (format == '-'))
    res = PyList_New(i);        /* list of triplets (size,supp,freq) */
  else {                        /* three columns for size/min/max */
    res     = PyList_New(3); if (!res) return NULL;
    cols[0] = PyList_New(i); if (!cols[0]) { clean1(res); return NULL; }
    PyList_SET_ITEM(res, 0, cols[0]);  /* sizes */
    cols[1] = PyList_New(i); if (!cols[1]) { clean1(res); return NULL; }
    PyList_SET_ITEM(res, 1, cols[1]);  /* support values */
    cols[2] = PyList_New(i); if (!cols[2]) { clean1(res); return NULL; }
    PyList_SET_ITEM(res, 2, cols[2]);  /* frequencies */
  }                             /* (create the return object) */
  if (!res) return NULL;        /* check for successful creation */
  for (i = 0, size = psp_min(psp); size <= psp_max(psp); size++) {
    min = psp_min4sz(psp,size); /* traverse the pattern sizes */
    max = psp_max4sz(psp,size); /* get range of support values */
    if (max < min) continue;    /* skip size with empty support range */
    z = PyInt_FromLong((long)size);  /* create objects for storing */
    s = PyFloat_FromDouble((double)min);
    m = PyFloat_FromDouble((double)max);
    if (!z || !s || !m) { clean3(z, s, m); e = -1; break; }
    if      (format == '#') {   /* if pattern spectrum is dictionary */
      t = PyTuple_New(2);       /* create a tuple as value */
      if (!t) { clean3(z, s, m); e = -1; break; }
      PyTuple_SET_ITEM(t, 0, s);/* create value as pair (min,max) */
      PyTuple_SET_ITEM(t, 1, m);/* and map size to support range */
      PyDict_SetItem(res, z, t); }
    else if (format == '=') {   /* if pattern spectrum is list */
      t = PyTuple_New(3);       /* create a result list element */
      if (!t) { clean3(z, s, m); e = -1; break; }
      PyTuple_SET_ITEM(t, 0, z);/* fill element with triplet */
      PyTuple_SET_ITEM(t, 1, s);/* (size, supp_min, supp_max) */
      PyTuple_SET_ITEM(t, 2, m);/* and set it in the result */
      PyList_SET_ITEM(res, i, t); }
    else {                      /* if result is three columns */
      PyList_SET_ITEM(cols[0], i, z);
      PyList_SET_ITEM(cols[1], i, s);
      PyList_SET_ITEM(cols[2], i, m);
    }                           /* set size and support range */
    i += 1;                     /* count the processed pattern size */
  }
  if (e) { clean1(res); return NULL; }
  return res;                   /* return created pattern spectrum */
}  /* psp_toPyObj() */

#endif  /* #ifndef CCN_OVERLAP .. #else .. */
/*--------------------------------------------------------------------*/

static int repinit (REPDATA *data, ISREPORT *isrep, CCHAR *report)
{                               /* --- initialize reporting */
  assert(data && isrep && report); /* check the function arguments */
  data->err = 0;                /* initialize the error indicator */
  if ((report[0] == '#')        /* if to get a pattern spectrum */
  ||  (report[0] == '=')        /* '#':     dictionary */
  ||  (report[0] == '-')        /* '=','-': list of triplets */
  ||  (report[0] == '|'))       /* '|':     three columns */
    return isr_addpsp(isrep, NULL);
  data->mode = ((report[0] == '(') || (report[0] == '['))
             ? *report++ : 0;   /* note mode and number of values */
  data->cnt  = (int)strlen(data->rep = report);
  data->res  = PyList_New(0);   /* create an empty output list */
  if (!data->res) return -1;    /* and set the reporting function */
  isr_setrepo(isrep, isr_toPyObj, data);
  return 0;                     /* return 'ok' */
}  /* repinit() */

/*--------------------------------------------------------------------*/

static int repterm (REPDATA *data, ISREPORT *isrep, CCHAR *report)
{                               /* --- terminate reporting */
  assert(data && isrep && report); /* check the function arguments */
  if (data->err) return data->err = -1;
  if ((report[0] == '#')        /* if to get a pattern spectrum */
  ||  (report[0] == '=')        /* '#':     dictionary */
  ||  (report[0] == '-')        /* '=','-': list of triplets */
  ||  (report[0] == '|')) {     /* '|':     three columns */
    data->res = psp_toPyObj(isr_getpsp(isrep), 1.0, report[0]);
    return data->err = (data->res) ? 0 : -1;
  }                             /* make Python pattern spectrum */
  return data->err;             /* return the error status */
}  /* repterm() */

/*--------------------------------------------------------------------*/

static void repfn (long cnt, void *data)
{                               /* --- progress reporting function */
  if ((cnt > *(long*)data) && ((cnt % 20) == 0))
    fprintf(stderr, "%10ld\b\b\b\b\b\b\b\b\b\b", *(long*)data = cnt);
}  /* repfn() */

/*--------------------------------------------------------------------*/
/* coconad (trains, target='c', width=0.003, supp=2,                  */
/*          zmin=2, zmax=None, report='a', mode='', algo='r',         */
/*          [mix=0.0,] border=None)                                   */
/*--------------------------------------------------------------------*/

static PyObject* py_coconad (PyObject *self,
                             PyObject *args, PyObject *kwds)
{                               /* --- run the CoCoNAD algorithm */
  char     *ckwds[] = { "trains", "target", "width", "supp",
                        "zmin", "zmax", "report", "algo", "mode",
  #ifndef CCN_OVERLAP           /* if standard (MIS) support */
                        "border", NULL };
  #else                         /* if influence map overlap support */
                        "mix", "border", NULL };
  #endif                        /* (additional mixing parameter) */
  CCHAR    *starg  = "c";       /* target type as a string */
  int      target  = ISR_CLOSED;/* target type identifier */
  double   width   =  0.003;    /* width of influence region */
  double   supp    =  2;        /* minimum support of an item set */
  long     zmin    =  2;        /* minimum size of an item set */
  long     zmax    = -1;        /* maximum size of an item set */
  char     *report = "a";       /* indicators of values to report */
  CCHAR    *salgo  = "";        /* algorithm as a string */
  #ifdef CCN_OVERLAP            /* if influence map overlap support */
  int      algo    = CCN_ISECT; /* CoCoNAD algorithm */
  double   mix     = 0.0;       /* support mixing parameter */
  #else                         /* if standard (MIS) support */
  int      algo    = CCN_REPO;  /* CoCoNAD algorithm variant */
  #endif                        /* (default algo. depends on support) */
  CCHAR    *smode  = "", *s;    /* operation mode/flags as a string */
  int      mode    = CCN_DEFAULT;  /* operation mode/flags */
  PyObject *border = NULL;      /* support border for filtering */
  int      r;                   /* result of function call */
  PyObject *trains;             /* (spike) train database */
  TRAINSET *trnset;             /* set of (spike) trains */
  ITEM     n;                   /* number of (spike) trains */
  ISREPORT *isrep;              /* item set reporter */
  COCONAD  *coconad;            /* coconad miner */
  REPDATA  data;                /* data for item set reporting */

  /* --- evaluate the function arguments --- */
  #ifndef CCN_OVERLAP           /* if standard (MIS) support */
  if (!PyArg_ParseTupleAndKeywords(args, kwds, "O|sddllsssO", ckwds,
         &trains, &starg, &width, &supp, &zmin, &zmax, &report,
         &salgo,  &smode, &border))
  #else                         /* if influence map overlap support */
  if (!PyArg_ParseTupleAndKeywords(args, kwds, "O|sddllsssdO", ckwds,
         &trains, &starg, &width, &supp, &zmin, &zmax, &report,
         &salgo,  &smode, &mix, &border))
  #endif                        /* additional mixing parameter */
    return NULL;                /* parse the function arguments */
  target = get_target(starg);   /* translate the target string */
  if (target < 0) return NULL;
  if (width  < 0) return ERR_VALUE("width must be >= 0");
  if (supp   < 0) supp = -supp; /* ensure non-negative support */
  if (zmin   < 1) return ERR_VALUE("zmin must be > 0");
  if (zmax   < 1) zmax = ITEM_MAX;  /* check the size range */
  if (zmax   < zmin) return ERR_VALUE("zmax must be >=  zmin");
  algo = get_coconad(salgo, target);
  if (algo   < 0) return NULL;  /* translate algorithm code */
  for (s = smode; *s; s++) {    /* traverse the mode characters */
    if      (*s == 'x') mode &= ~CCN_PERFECT;
    else if (*s == '0') mode |=  CCN_REPOLIST;
    else if (*s == '1') mode |=  CCN_REPOTREE;
    else if (*s == '2') mode |=  CCN_REPOCHAIN;
    #ifndef CCN_OVERLAP         /* if standard (MIS) support */
    else if (*s == 'p') mode &= ~CCN_REORDER;
    #endif                      /* adapt the operation mode */
  }                             /* based on letters in mode string */

  /* --- create train set --- */
  sig_install();                /* install the signal handler */
  trnset = tns_fromPyObj(trains);
  if (!trnset) return ERR_RTN();/* create a train set */
  if (zmin > ITEM_MAX) zmin = ITEM_MAX;
  if (zmax > ITEM_MAX) zmax = ITEM_MAX;
  coconad = ccn_create(target, width, supp,
  #ifdef CCN_OVERLAP            /* if influence map overlap support */
                       mix,     /* add. support mixing parameter */
  #endif
                       (ITEM)zmin, (ITEM)zmax, algo, mode);
  if (!coconad) {               /* create a coconad miner */
    tns_delete(trnset,  1); return ERR_MEM(); }
  if (ccn_data(coconad, trnset, 0) != 0) {
    ccn_delete(coconad, 1); return ERR_MEM(); }

  /* --- create item set reporter --- */
  n = tns_cnt(trnset);          /* get the number of trains */
  #ifndef CCN_OVERLAP           /* if standard (MIS) support */
  if (target & CCN_SEQUENCE)    /* if sequences (with poss. reps.) */
    n = tns_maxseq(trnset, width) +1;
  #endif                        /* find max. possible sequence len. */
  isrep = isr_createx(tns_base(trnset), n);
  if (!isrep                    /* create an item set reporter */
  ||  (ccn_report(coconad, isrep) != 0)) {
    ccn_delete(coconad, 1); return ERR_MEM(); }
  if (isr_pyborder(isrep, border) == NULL) {
    ccn_delete(coconad, 1); return ERR_RTN(); }
  if ((repinit(&data, isrep, report) != 0)
  ||  (isr_setup(isrep) < 0)) { /* set up the item set reporter */
    ccn_delete(coconad, 1); return ERR_MEM(); }

  /* --- frequent item set mining --- */
  r = ccn_mine(coconad);
  if (r >= 0) r = repterm(&data, isrep, report);

  /* --- clean up --- */
  ccn_delete(coconad, 1);       /* delete the coconad miner */
  if (sig_aborted()) { sig_abort(0);
    clean1(data.res); PyErr_SetInterrupt(); return ERR_ABORT(); }
  if (r < 0) { clean1(data.res); return ERR_MEM(); }
  sig_remove();                 /* remove the signal handler */
  return data.res;              /* return the created result */
}  /* py_coconad() */

/*--------------------------------------------------------------------*/
/* cocojim (trains, target='c', width=0.003, supp=2,                  */
/*          simcov='j', simps=[], sim=0.01, beg=-oo, end=+oo,         */
/*          zmin=2, zmax=None, report='a', algo='i',                  */
/*          mode='', border=None)                                     */
/*--------------------------------------------------------------------*/
#ifdef CCN_OVERLAP

static PyObject* py_cocojim (PyObject *self,
                             PyObject *args, PyObject *kwds)
{                               /* --- run the CoCoNAD algorithm */
  char     *ckwds[] = { "trains", "target", "width", "supp",
                        "simcov", "simps", "sim", "beg", "end",
                        "zmin", "zmax", "report", "algo",
                        "mode", "border", NULL };
  CCHAR    *starg  = "c";       /* target type as a string */
  int      target  = ISR_CLOSED;/* target type identifier */
  double   width   =  0.003;    /* width of influence region */
  double   supp    =  2;        /* minimum support of an item set */
  char     *scov   = "j";       /* item cover similarity as a string */
  int      simcov  = -1;        /* and as an identifier */
  PyObject *spars  = NULL;      /* similarity measure parameters */
  double   simps[8]= { 1, 0, 0, 0,  0, -1, 1, 0 };
  double   sim     = 0.01;      /* minimum similarity measure */
  double   beg     = -INFINITY; /* beginning of event range */
  double   end     = +INFINITY; /* end       of event range */
  long     zmin    =  2;        /* minimum size of an item set */
  long     zmax    = -1;        /* maximum size of an item set */
  char     *report = "a";       /* indicators of values to report */
  CCHAR    *salgo  = "i";       /* algorithm as a string */
  int      algo    = CCN_ISECT; /* CoCoNAD algorithm */
  char     *smode  = "", *s;    /* search mode */
  int      mode    = CCN_DEFAULT;
  PyObject *border = NULL;      /* support border for filtering */
  int      r;                   /* result of function call */
  PyObject *trains;             /* (spike) train database */
  TRAINSET *trnset;             /* set of (spike) trains */
  ISREPORT *isrep;              /* item set reporter */
  COCONAD  *coconad;            /* coconad miner */
  REPDATA  data;                /* data for item set reporting */

  /* --- evaluate the function arguments --- */
  if (!PyArg_ParseTupleAndKeywords(args, kwds, "O|sddsOdddllsssO",
         ckwds, &trains, &starg, &width, &supp,
         &scov, &spars, &sim, &beg, &end, &zmin, &zmax,
         &report, &salgo, &smode, &border))
    return NULL;                /* parse the function arguments */
  target = get_target(starg);   /* translate the target string */
  if (target < 0) return NULL;
  if (width  < 0) return ERR_VALUE("width must be >= 0");
  if (supp   < 0) supp = -supp; /* ensure non-negative support */
  simcov = get_sim(scov);       /* translate the similarity */
  if (simcov < 0) return NULL;  /* measure string */
  if ((simcov == CCN_GENERIC) && !sim_fromPyObj(simps, spars))
    return NULL;                /* get similarity measure parameters */
  if (sim    < 0) return ERR_VALUE("sim must be >= 0");
  if (zmin   < 1) return ERR_VALUE("zmin must be > 0");
  if (zmax   < 1) zmax = ITEM_MAX;  /* check the size range */
  if (zmax   < zmin) return ERR_VALUE("zmax must be >=  zmin");
  algo = get_coconad(salgo, target);
  if (algo   < 0) return NULL;  /* translate algorithm code */
  for (s = smode; *s; s++) {    /* traverse the mode characters */
    if      (*s == 'x') mode &= ~CCN_PERFECT;
    else if (*s == 'e') mode |=  CCN_SIMSUPP;
    else if (*s == '0') mode |=  CCN_REPOLIST;
  }                             /* adapt the operation mode */

  /* --- create train set --- */
  sig_install();                /* install the signal handler */
  trnset = tns_fromPyObj(trains);
  if (!trnset) return ERR_RTN();/* create a train set */
  if (zmin > ITEM_MAX) zmin = ITEM_MAX;
  if (zmax > ITEM_MAX) zmax = ITEM_MAX;
  coconad = ccn_create(target, width, supp, 0,
                       (ITEM)zmin, (ITEM)zmax, algo, mode);
  if (!coconad) {               /* create a coconad miner */
    tns_delete(trnset,  1); return ERR_MEM(); }
  if (ccn_data(coconad, trnset, 0) != 0) {
    ccn_delete(coconad, 1); return ERR_MEM(); }

  /* --- create item set reporter --- */
  isrep = isr_create(tns_base(trnset));
  if (!isrep                    /* create an item set reporter */
  ||  (ccn_report(coconad, isrep) != 0)) {
    ccn_delete(coconad, 1); return ERR_MEM(); }
  if (isr_pyborder(isrep, border) == NULL) {
    ccn_delete(coconad, 1); return ERR_RTN(); }
  if ((repinit(&data, isrep, report) != 0)
  ||  (isr_setup(isrep) < 0)) { /* set up the item set reporter */
    ccn_delete(coconad, 1); return ERR_MEM(); }

  /* --- frequent item set mining --- */
  if (simcov <= CCN_NONE)       /* if no similarity measure is given */
    r = ccn_mine(coconad);      /* call the standard mining function, */
  else                          /* otherwise call cocojim function */
    r = ccn_jim(coconad, simcov, simps, sim*0.01, beg, end);
  if (r >= 0) r = repterm(&data, isrep, report);

  /* --- clean up --- */
  ccn_delete(coconad, 1);       /* delete the coconad miner */
  if (sig_aborted()) { sig_abort(0);
    clean1(data.res); PyErr_SetInterrupt(); return ERR_ABORT(); }
  if (r < 0) { clean1(data.res); return ERR_MEM(); }
  sig_remove();                 /* remove the signal handler */
  return data.res;              /* return the created result */
}  /* py_cocojim() */

#endif
/*--------------------------------------------------------------------*/
/* genpsp (trains, target='c', width=0.003, supp=2,                   */
/*         zmin=2, zmax=None, report='#', algo='r', mode='',[mix=0.0],*/
/*         cnt=1000, beg=-oo, end=+oo, surr='p', rand='u',            */
/*         sigma=0.005, delta=0.03, seed=0, cpus=0)                   */
/*--------------------------------------------------------------------*/

static PyObject* py_genpsp (PyObject *self,
                            PyObject *args, PyObject *kwds)
{                               /* --- generate a pattern spectrum */
  char      *ckwds[] = { "trains", "target", "width", "supp",
                         "zmin", "zmax", "report", "algo", "mode",
  #ifdef CCN_OVERLAP            /* if influence map overlap support */
                         "mix", /* mixing parameter for standard */
  #endif                        /* and influence map overlap support */
                         "cnt", "beg", "end", "surr", "rand",
                         "sigma", "delta", "seed", "cpus", NULL };
  CCHAR    *starg  = "c";       /* target type as a string */
  int      target  = ISR_CLOSED;/* target type identifier */
  double   width   =  0.003;    /* width of influence region */
  double   supp    =  2;        /* minimum support of an item set */
  long     zmin    =  2;        /* minimum size of an item set */
  long     zmax    = -1;        /* maximum size of an item set */
  char     *report = "#";       /* indicators of reporting format */
  CCHAR    *salgo  = "";        /* algorithm as a string */
  #ifdef CCN_OVERLAP            /* if influence map overlap support */
  int      algo    = CCN_ISECT; /* CoCoNAD algorithm */
  double   mix     = 0.0;       /* support mixing parameter */
  #else                         /* if standard (MIS) support */
  int      algo    = CCN_REPO;  /* CoCoNAD algorithm variant */
  #endif                        /* (default algo. depends on support) */
  CCHAR    *smode  = "", *s;    /* operation mode/flags as a string */
  int      mode    = CCN_DEFAULT;  /* operation mode/flags */
  long     cnt     = 1000;      /* number of data sets to generate */
  double   beg     = -INFINITY; /* beginning of event range */
  double   end     = +INFINITY; /* end       of event range */
  CCHAR    *ssurr  = "p";       /* surrogate method as a string */
  int      surr    =  5;        /* surrogate method identifier */
  CCHAR    *srand  = "r";       /* random function as a string */
  int      rand    = RNG_RECT;  /* random function identifier */
  double   sigma   =  0.005;    /* parameter of random function */
  double   delta   =  0.03;     /* limit for localized swaps */
  long     seed    =  0;        /* seed for random number generator */
  int      cpus    =  0;        /* number of cpus */
  long     done    =  0;        /* number of completed surrogates */
  PATSPEC  *psp    = NULL;      /* created pattern spectrum (C)*/
  PyObject *pypsp  = NULL;      /* created pattern spectrum (Python) */
  PyObject *trains;             /* (spike) train database (Python) */
  TRAINSET *trnset;             /* (spike) train database (C) */

  /* --- evaluate the function arguments --- */
  #ifndef CCN_OVERLAP           /* if standard (MIS) support */
  if (!PyArg_ParseTupleAndKeywords(args, kwds, "O|sddllssslddssddli",
         ckwds, &trains, &starg, &width, &supp, &zmin, &zmax, &report,
         &salgo, &smode, &cnt, &beg, &end, &ssurr, &srand, &sigma,
         &delta, &seed, &cpus)) /* parse the function arguments */
  #else                         /* if influence map overlap support */
  if (!PyArg_ParseTupleAndKeywords(args, kwds, "O|sddllsssdlddssddli",
         ckwds, &trains, &starg, &width, &supp, &zmin, &zmax, &report,
         &salgo, &smode, &mix, &cnt, &beg, &end, &ssurr, &srand, &sigma,
         &delta, &seed, &cpus)) /* parse the function arguments */
  #endif                        /* (additional mixing parameter) */
    return NULL;
  target = get_target(starg);   /* translate the target string */
  if (target < 0)  return NULL;
  if (width  < 0)  return ERR_VALUE("width must be >= 0");
  if (supp   < 0)  supp = -supp;    /* ensure non-negative support */
  if (zmin   < 1)  return ERR_VALUE("zmin must be > 0");
  if (zmax   < 1)  zmax = ITEM_MAX; /* check the size range */
  if (zmax < zmin) return ERR_VALUE("zmax must be >= zmin");
  algo = get_coconad(salgo, target);
  if (algo  <  0)  return NULL; /* translate algorithm variant */
  for (s = smode; *s; s++) {    /* traverse the mode characters */
    if      (*s == 'x') mode &= ~CCN_PERFECT;
    #ifndef CCN_OVERLAP         /* if standard (MIS) support */
    else if (*s == 'p') mode &= ~CCN_REORDER;
    #endif                      /* adapt the operation mode */
  }
  if (cnt   <= 0)  cnt = 1;     /* check the number of data sets */
  surr   = get_surr(ssurr);     /* translate the surrogate string */
  if (surr  <  0)  return NULL;
  if (surr  == 0)  cnt = 1;     /* only one surrogate for identity */
  rand   = get_rand(srand);     /* translate random function string */
  if (rand  <  0)  return NULL;
  if (sigma <  0)  return ERR_VALUE("sigma must not be negative");
  if (delta <= 0)  return ERR_VALUE("delta must be positive");

  /* --- generate pattern spectrum --- */
  sig_install();                /* install the signal handler */
  trnset = tns_fromPyObj(trains);
  if (!trnset) return ERR_RTN();/* create a train set */
  if (zmin > ITEM_MAX) zmin = ITEM_MAX;
  if (zmax > ITEM_MAX) zmax = ITEM_MAX;
  #ifndef CCN_OVERLAP           /* if standard (MIS) support */
  psp = ccn_genpsp(trnset, target, width, (SUPP)supp,
  #else                         /* if influence map overlap support */
  psp = ccn_genpsp(trnset, target, width, (SUPP)supp, mix,
  #endif                        /* additional mixing parameter */
                   (ITEM)zmin, (ITEM)zmax, algo, mode,
                   (size_t)cnt, beg, end, surr, rand,
                   sigma, delta, seed, cpus, repfn, &done);
  if (psp) { pypsp = psp_toPyObj(psp, 1.0/(double)cnt, report[0]);
             psp_delete(psp); } /* generate a pattern spectrum */
  tns_delete(trnset, 1);        /* delete the created train set */
  if (sig_aborted()) { sig_abort(0);
    clean1(pypsp); PyErr_SetInterrupt(); return ERR_ABORT(); }
  if (!pypsp) return ERR_MEM(); /* check for an error */
  sig_remove();                 /* remove the signal handler */
  return pypsp;                 /* return created pattern spectrum */
}  /* py_genpsp() */

/*--------------------------------------------------------------------*/
/* estpsp (trains, target="a", width=0.003, supp=2, zmin=2, zmax=None,*/
/*         report='#', equiv=10000, alpha=0.5, smpls=1000, seed=0)    */
/*--------------------------------------------------------------------*/
#ifndef CCN_OVERLAP

static PyObject* py_estpsp (PyObject *self,
                            PyObject *args, PyObject *kwds)
{                               /* --- generate a pattern spectrum */
  char    *ckwds[] = { "trains", "target", "width", "supp",
                       "zmin", "zmax", "report", "equiv",
                       "alpha", "smpls", "seed", NULL };
  CCHAR    *starg  = "a";       /* target type as a string */
  int      target  = ISR_ALL;   /* target type identifier */
  double   width   =  0.003;    /* width of influence region */
  double   supp    =  2;        /* minimum support of an item set */
  long     zmin    =  2;        /* minimum size of an item set */
  long     zmax    = -1;        /* maximum size of an item set */
  char     *report = "#";       /* indicators of reporting format */
  long     equiv   = 10000;     /* equivalent number of surrogates */
  double   alpha   =  0.5;      /* probability dispersion factor */
  long     smpls   = 1000;      /* number of samples per set size */
  long     seed    =  0;        /* seed for random number generator */
  PATSPEC  *psp    = NULL;      /* created pattern spectrum */
  PyObject *pypsp  = NULL;      /* created Python pattern spectrum */
  PyObject *trains;             /* (spike) train database (Python) */
  TRAINSET *trnset;             /* (spike) train database (C) */

  /* --- evaluate the function arguments --- */
  if (!PyArg_ParseTupleAndKeywords(args, kwds, "O|sddllsldll", ckwds,
        &trains, &starg, &width, &supp, &zmin, &zmax, &report,
        &equiv, &alpha, &smpls, &seed))
    return NULL;                /* parse the function arguments */
  target = get_target(starg);   /* translate the target string */
  if (target < 0) return NULL;  /* and check for a valid target */
  if (width <= 0) return ERR_VALUE("width must be > 0");
  if (supp  <  0) supp = -supp;    /* ensure non-negative support */
  if (zmin  <  1) return ERR_VALUE("zmin must be > 0");
  if (zmax  <  1) zmax = ITEM_MAX; /* check the size range */
  if (zmax  <  zmin) return ERR_VALUE("zmax must be >= zmin");
  if (equiv <= 0) equiv = 1; /* check the number of data sets */
  if (smpls <= 0) return ERR_VALUE("smpls must be positive");
  if (seed  == 0) seed = (long)time(NULL);

  /* --- estimate pattern spectrum --- */
  sig_install();                /* install the signal handler */
  trnset = tns_fromPyObj(trains);
  if (!trnset) return ERR_RTN();/* create a train set and */
  tns_sort(trnset, +1);         /* sort the trains by times */
  if (zmin > ITEM_MAX) zmin = ITEM_MAX;
  if (zmax > ITEM_MAX) zmax = ITEM_MAX;
  psp = ccn_estpsp(trnset, target, width, (SUPP)supp,
                   (ITEM)zmin, (ITEM)zmax,
                   (size_t)equiv, alpha, (size_t)smpls, seed);
  if (psp) { pypsp = psp_toPyObj(psp, 1/(double)equiv, report[0]);
             psp_delete(psp); } /* estimate a pattern spectrum */
  tns_delete(trnset, 1);        /* delete the train set */
  if (sig_aborted()) { sig_abort(0);
    clean1(pypsp); PyErr_SetInterrupt(); return ERR_ABORT(); }
  if (!pypsp) return ERR_MEM(); /* check for an error */
  sig_remove();                 /* remove the signal handler */
  return pypsp;                 /* the created pattern spectrum */
}  /* py_estpsp() */

#endif
/*--------------------------------------------------------------------*/
/* psp2bdr (psp)                                                      */
/*--------------------------------------------------------------------*/

static PyObject* py_psp2bdr (PyObject *self,
                             PyObject *args, PyObject *kwds)
{                               /* --- extract a decision border */
  char     *ckwds[] = { "psp", NULL };
  PyObject *psp;                /* pattern spectrum */
  PyObject *ei;                 /* pattern spectrum element iterator */
  PyObject *elem;               /* to traverse the elements */
  PyObject *size;               /* to traverse the sizes */
  PyObject *supp;               /* to traverse the support values */
  ITEM     z, zmax;             /* (maximum) pattern size */
  RSUPP    s, *bdr;             /* pattern support, decision border */
  PyObject *pybdr;              /* Python decision border */

  /* --- evaluate the function arguments --- */
  if (!PyArg_ParseTupleAndKeywords(args, kwds, "O", ckwds, &psp))
    return NULL;                /* parse the function arguments */

  /* --- find maximum size --- */
  zmax = 2;                     /* initialize the maximum size */
  ei   = PyObject_GetIter(psp); /* get an iterator for the trains */
  if (!ei) return ERR_TYPE("pattern spectrum must be iterable");
  while ((elem = PyIter_Next(ei)) != NULL) {
    #ifdef CCN_OVERLAP          /* if overlap-based support */
    if (!PySequence_Check(elem) /* elements must be sequences */
    ||  (PySequence_Length(elem) < 3)) { clean2(elem, ei);
      return ERR_TYPE("pattern spectrum elements "
                      "must have length >= 3"); }
    #else                       /* if standard support */
    if (!PySequence_Check(elem) /* elements must be sequences */
    ||  (PySequence_Length(elem) < 2)) { clean2(elem, ei);
      return ERR_TYPE("pattern spectrum elements "
                      "must have length >= 2"); }
    #endif                      /* check pattern spectrum element */
    size = PySequence_GetItem(elem, 0);
    if (!size) { clean2(elem, ei); return NULL; }
    if      (PyLong_Check(size))/* if size is a long integer */
      z = (ITEM)PyLong_AsLong(size);
    else if (PyInt_Check(size)) /* if size is an integer */
      z = (ITEM)PyInt_AsLong(size);
    else if (PyFloat_Check(size))  /* if size is a float */
      z = (ITEM)PyFloat_AsDouble(size);
    else z = 0;                 /* interpret all else as 0 */
    Py_DECREF(size);            /* drop the size reference */
    Py_DECREF(elem);            /* drop the element reference */
    if (z >= zmax) zmax = z;    /* update the maximum size */
  }
  Py_DECREF(ei);                /* drop the element iterator */

  /* --- find decision border --- */
  bdr = (RSUPP*)calloc((size_t)(zmax+1), sizeof(RSUPP));
  if (!bdr) return ERR_MEM();   /* create a decision border */
  ei = PyObject_GetIter(psp);   /* get an iterator for the trains */
  if (!ei) return ERR_TYPE("pattern spectrum must be iterable");
  while ((elem = PyIter_Next(ei)) != NULL) {
    #ifdef CCN_OVERLAP          /* if overlap-based support */
    if (!PySequence_Check(elem) /* elements must be sequences */
    ||  (PySequence_Length(elem) < 3)) { clean2(elem, ei);
      return ERR_TYPE("pattern spectrum elements "
                      "must have length >= 3"); }
    #else                       /* if standard support */
    if (!PySequence_Check(elem) /* elements must be sequences */
    ||  (PySequence_Length(elem) < 2)) { clean2(elem, ei);
      return ERR_TYPE("pattern spectrum elements "
                      "must have length >= 2"); }
    #endif                      /* check pattern spectrum element */
    size = PySequence_GetItem(elem, 0);
    if (!size) { clean2(elem, ei); return NULL; }
    if      (PyLong_Check(size))/* if size is a long integer */
      z = (ITEM)PyLong_AsLong(size);
    else if (PyInt_Check(size)) /* if size is an integer */
      z = (ITEM)PyInt_AsLong(size);
    else if (PyFloat_Check(size))  /* if size is a float */
      z = (ITEM)PyFloat_AsDouble(size);
    else z = 0;                 /* interpret all else as 0 */
    Py_DECREF(size);            /* drop the size reference */
    #ifdef CCN_OVERLAP          /* if overlap-based support */
    supp = PySequence_GetItem(elem, 2);
    #else                       /* if normal support (integer) */
    supp = PySequence_GetItem(elem, 1);
    #endif                      /* get the support value */
    if (!supp) { clean3(size, elem, ei); return NULL; }
    if      (PyLong_Check(size))/* if support is a long integer */
      s = (RSUPP)PyLong_AsLong(supp);
    else if (PyInt_Check(supp)) /* if support is an integer */
      s = (RSUPP)PyInt_AsLong(supp);
    else if (PyFloat_Check(supp))  /* if support is a float */
      s = (RSUPP)PyFloat_AsDouble(supp);
    else s = 0;                 /* interpret all else as 0 */
    Py_DECREF(supp);            /* drop the support reference */
    Py_DECREF(elem);            /* drop the element reference */
    if (s > bdr[z]) bdr[z] = s; /* update the decision border */
  }                             /* (maximum support per size) */

  /* --- build decision border --- */
  for (z = zmax; --z >= 0; )    /* ensure monotone border */
    if (bdr[z+1] > bdr[z]) bdr[z] = bdr[z+1];
  pybdr = PyList_New((Py_ssize_t)(zmax+1));
  if (!pybdr) return ERR_MEM(); /* create a border list */
  supp  = PyFloat_FromDouble(INFINITY);
  PyList_SET_ITEM(pybdr, 0, supp);
  PyList_SET_ITEM(pybdr, 1, supp);
  for (z = 2; z <= zmax; z++) { /* traverse the pattern sizes */
    #ifdef CCN_OVERLAP          /* if overlap-based support */
    supp = PyFloat_FromDouble((double)bdr[z]);
    #else                       /* if standard support (integer) */
    supp = PyInt_FromLong((long)bdr[z]+1);
    #endif                      /* get support as Python object */
    if (!supp) { clean1(pybdr); return ERR_MEM(); }
    PyList_SET_ITEM(pybdr, z, supp);
  }                             /* set minimum support value */
  return pybdr;                 /* return created decision border */
}  /* py_psp2bdr() */

/*--------------------------------------------------------------------*/
/* patred (pats, method='S', border=None, addis=True)                 */
/*--------------------------------------------------------------------*/

static PyObject* py_patred (PyObject *self,
                            PyObject *args, PyObject *kwds)
{                               /* --- extract a decision border */
  char     *ckwds[] = { "pats", "method", "border", "addis", NULL };
  PyObject *pats   = NULL;      /* pattern set to reduce */
  CCHAR    *smeth  = "S";       /* pattern set reduction method */
  int      method  = PSR_COVER1;/* pattern set reduction method */
  PyObject *border = NULL;      /* support border for filtering */
  char     addis   = -1;        /* whether to add intersections */
  IDMAP    *map;                /* item to identifier map */
  PATSET   *patset;             /* internal pattern set */
  size_t   i, n;                /* loop variable, number of patterns */
  size_t   k, z;                /* (maximal) size of a pattern */
  PyObject *pi;                 /* pattern iterator */
  PyObject *pat;                /* to traverse the patterns */
  PyObject *iset;               /* to traverse the item sets */
  PyObject *ii;                 /* item iterator */
  PyObject *item;               /* to traverse the items */
  PyObject *supp;               /* to traverse the support values */
  PyObject *red;                /* reduced pattern set */
  int      dict;                /* whether pattern set is dictionary */

  /* --- evaluate the function arguments --- */
  if (!PyArg_ParseTupleAndKeywords(args, kwds, "O|sOb", ckwds,
        &pats, &smeth, &border, &addis))
    return NULL;                /* parse the function arguments */
  method = get_red(smeth);      /* translate the method string */
  if (method < 0) return NULL;  /* into an identifier/offset */

  /* --- find maximum size and item count --- */
  sig_install();                /* install the signal handler */
  n = i = 0; k = 1;             /* init. counters and max. size */
  dict = PyDict_Check(pats);    /* check whether pattern set is dict. */
  pi = PyObject_GetIter(pats);  /* get a pattern iterator */
  if (!pi) return ERR_TYPE("pattern set must be iterable");
  while ((pat = PyIter_Next(pi)) != NULL) {
    if (dict) iset = pat;       /* if dictionary, key is item set */
    else {                      /* if some other iterable, */
      if (!PySequence_Check(pat)/* patterns must be sequences */
      ||  (PySequence_Length(pat) < 2)) { clean2(pat, pi);
        return ERR_TYPE("patterns in non-dictionary must be pairs "
                        "(sequences with length >= 2)"); }
      iset = PySequence_GetItem(pat, 0);
      if (!iset) { clean2(pat, pi); return NULL; }
      Py_DECREF(pat);           /* get the item set and */
    }                           /* drop the pattern reference */
    if      (PySequence_Check(iset))     /* if sequence, */
      z = (size_t)PySequence_Size(iset); /* get its length */
    else if (PyAnySet_Check(iset))       /* if set or frozenset, */
      z = (size_t)PySet_Size(iset);      /* get its size */
    else {                      /* if any other iterable object */
      ii = PyObject_GetIter(iset);  /* check for an iterable */
      if (!ii) { clean2(iset, pi);  /* (list, tuple, set etc.) */
        return ERR_TYPE("item set of a pattern must be iterable"); }
      for (z = 0; (item = PyIter_Next(ii)) != NULL; z++)
        Py_DECREF(item);        /* count the number of items and */
      Py_DECREF(ii);            /* drop the iterartor reference */
      if (PyErr_Occurred()) { clean2(iset, pi); return NULL; }
    }
    Py_DECREF(iset);            /* drop the item set reference */
    if (z > k) k = z;           /* update maximum size and extent */
    i += z; n += 1;             /* and count the pattern */
    if (sig_aborted()) break;   /* check for user abort */
  }
  Py_DECREF(pi);                /* drop the pattern iterator */
  if (PyErr_Occurred()) return NULL;
  if (sig_aborted()) { sig_abort(0);
    PyErr_SetInterrupt(); return ERR_ABORT(); }
  /* n: number of patterns */
  /* k: maximum size of a pattern */
  /* i: total number of item istances (extent) */

  /* --- create pattern set --- */
  z = (n > 255) ? n : 255;      /* compute identifier map size */
  map = idm_create(z, 0, hashitem, cmpitems, NULL, NULL);
  if (!map) return ERR_MEM();   /* create item map and pattern set */
  patset = psr_create(n, (ITEM)k, i, map);
  if (!patset) { idm_delete(map); return ERR_MEM(); }
  if (border) {                 /* if a decision border is given */
    if (!PySequence_Check(border)) { psr_delete(patset, 1);
      return ERR_TYPE("border must be a sequence"); }
    i = (size_t)PySequence_Length(border);
    if (i > k+1) i = k+1;       /* limit border to pattern size */
    for (z = 2; z < i; z++) {   /* traverse the border */
      supp = PySequence_GetItem(border, (Py_ssize_t)z);
      if (!supp) { psr_delete(patset, 1); return NULL; }
      if      (PyLong_Check(supp))  /* if support is a long integer */
        psr_setbdr(patset, (ITEM)z, (RSUPP)PyLong_AsLong(supp));
      else if (PyInt_Check(supp))   /* if support is an integer */
        psr_setbdr(patset, (ITEM)z, (RSUPP)PyInt_AsLong(supp));
      else if (PyFloat_Check(supp)) /* if support is a float */
        psr_setbdr(patset, (ITEM)z, (RSUPP)PyFloat_AsDouble(supp));
      Py_DECREF(supp);          /* drop the support reference */
    }                           /* (fill the decision border */
  }                             /*  with minimum support values) */
  if (sig_aborted()) { sig_abort(0);
    psr_delete(patset, 1); PyErr_SetInterrupt(); return ERR_ABORT(); }

  /* --- collect patterns --- */
  pi = PyObject_GetIter(pats);  /* get a pattern iterator */
  if (!pi) { psr_delete(patset, 1);
    return ERR_TYPE("pattern set must be iterable"); }
  for (i = 0; (pat = PyIter_Next(pi)) != NULL; i++) {
    psr_addorig(patset, pat);   /* note the input pattern */
    if (dict) {                 /* if dictionary, key is item set */
      iset = pat;               /* that is mapped to the support */
      supp = PyDict_GetItem(pats, iset); Py_INCREF(supp); }
    else {                      /* if some other iterable, */
      if (!PySequence_Check(pat)/* patterns must be sequences */
      ||  (PySequence_Length(pat) < 2)) {
        clean2(pat, pi); psr_delete(patset, 1);
        return ERR_TYPE("patterns in non-dictionary must be pairs "
                        "(sequences with length >= 2)"); }
      iset = PySequence_GetItem(pat, 0);
      if (!iset) { clean2(pat, pi); psr_delete(patset,1); return NULL; }
      supp = PySequence_GetItem(pat, 1);
      Py_DECREF(pat);           /* get item set and its support */
      if (!supp) { clean2(iset,pi); psr_delete(patset,1); return NULL; }
    }                           /* drop the pattern reference */
    ii = PyObject_GetIter(iset);/* check for an iterable */
    if (!ii) { clean3(supp, iset, pi); psr_delete(patset, 1);
      return ERR_TYPE("item set of a pattern must be an iterable"); }
    Py_DECREF(iset);            /* drop the item set reference */
    for (z = 0; (item = PyIter_Next(ii)) != NULL; z++) {
      if (psr_additem(patset, &item) != 0) { clean4(item, supp, ii, pi);
        psr_delete(patset, 1); return ERR_MEM(); }
      Py_DECREF(item);          /* add item to current pattern */
    }                           /* and drop the item reference */
    Py_DECREF(ii);              /* drop the item iterator */
    if      (PyInt_Check(supp)) /* get the pattern support */
      psr_addsupp(patset, (RSUPP)PyInt_AsLong(supp));
    else if (PyLong_Check(supp))
      psr_addsupp(patset, (RSUPP)PyLong_AsLong(supp));
    else if (PyFloat_Check(supp))
      psr_addsupp(patset, (RSUPP)PyFloat_AsDouble(supp));
    else { clean2(supp, pi); psr_delete(patset, 1);
      return ERR_TYPE("pattern support must be a number"); }
    Py_DECREF(supp);            /* drop the support reference */
    if (sig_aborted()) break;   /* check for user abort */
  }
  Py_DECREF(pi);                /* drop the pattern iterator */
  if (PyErr_Occurred()) { psr_delete(patset, 1); return NULL; }
  if (sig_aborted()) { sig_abort(0); psr_delete(patset, 1);
    PyErr_SetInterrupt(); return ERR_ABORT(); }

  /* --- pattern set reduction --- */
  k = psr_reduce(patset, method, addis);
  if (sig_aborted()) { sig_abort(0); psr_delete(patset, 1);
    PyErr_SetInterrupt(); return ERR_ABORT(); }
  if (dict) {                   /* if patterns in dictionary */
    red = PyDict_New();         /* create reduced pattern set */
    if (!red) { psr_delete(patset, 1); return ERR_MEM(); }
    for (i = 0; i < n; i++) {   /* traverse the patterns */
      pat = psr_getorig(patset, i);
      if (!pat) continue;       /* skip removed patterns */
      supp = PyDict_GetItem(pats, pat);
      if (!supp) { clean1(red); psr_delete(patset, 1); return NULL; }
      if (PyDict_SetItem(red, pat, supp) != 0) {
        clean1(red); psr_delete(patset, 1); return ERR_MEM(); }
      if (sig_aborted()) break; /* check for user abort */
    } }                         /* (return pattern set in same form) */
  else {                        /* if patterns in other iterable */
    red = PyList_New((Py_ssize_t)k);
    if (!red) { psr_delete(patset, 1); return ERR_MEM(); }
    for (i = k = 0; i < n; i++){/* traverse the patterns */
      pat = psr_getorig(patset, i);
      if (!pat) continue;       /* skip removed patterns */
      Py_INCREF(pat);           /* add pair as taken from pattern */
      PyList_SET_ITEM(red, (Py_ssize_t)k, pat);
      k += 1;                   /* count the output pattern */
      if (sig_aborted()) break; /* check for user abort */
    }
  }
  psr_delete(patset, 1);        /* clean up the pattern set */
  if (sig_aborted()) { sig_abort(0);
    clean1(pat); PyErr_SetInterrupt(); return ERR_ABORT(); }
  sig_remove();                 /* remove the signal handler */
  return red;                   /* return the reduced pattern set */
}  /* py_patred() */

/*--------------------------------------------------------------------*/
/* Python Function List                                               */
/*--------------------------------------------------------------------*/

static PyMethodDef coco_methods[] = {
  { "coconad", (PyCFunction)py_coconad, METH_VARARGS|METH_KEYWORDS,
    "coconad (trains, target='c', width=0.003, supp=2,\n"
    #ifdef CCN_OVERLAP
    "         zmin=2, zmax=None, report='a', algo='i', mode=''\n"
    "         mix=0.0, border=None)\n"
    #else
    "         zmin=2, zmax=None, report='a', algo='r', mode=''\n"
    "         border=None)\n"
    #endif
    "COntinuous time ClOsed Neuron Assembly Detection.\n"
    "trains  (spike) train database to mine         (mandatory)\n"
    "        The train database must be an iterable "
            "(list, tuple etc.).\n"
    "        If it is a dictionary, "
            "the keys must be items/event types\n"
    "        that are mapped to iterables of event/spike times.\n"
    "        Otherwise each database element must be a pair, "
            "consisting\n"
    "        of a item/event type and an iterable "
            "of event/spike times.\n"
    "        Each item identifier must be a hashable object.\n"
    "target  type of frequent item sets to find     (default: c)\n"
    "        a/s  all/sets   all     frequent item sets\n"
    "        c    closed     closed  frequent item sets\n"
    "        m    maximal    maximal frequent item sets\n"
    #ifndef CCN_OVERLAP
    "        ap   perms      all     frequent item permutations\n"
    "        cp   clsperms   closed  frequent item permutations\n"
    "        mp   maxperms   maximal frequent item permutations\n"
    "        as   seqs       all     frequent item sequences\n"
    "        cs   clsseqs    closed  frequent item sequences\n"
    "        ms   maxseqs    maximal frequent item sequences\n"
    #endif
    "width   width of time window/influence region  (default: 0.003)\n"
    "supp    minimum support of an item set         (default: 2)\n"
    "zmin    minimum number of items per item set   (default: 2)\n"
    "zmax    maximum number of items per item set   "
            "(default: no limit)\n"
    "report  values to report with an item set      (default: a)\n"
    "        a    absolute item set support (number of coincidences)\n"
    "        (    combine values in a tuple (must be first character)\n"
    "        [    combine values in a list  (must be first character)\n"
    "        #    pattern spectrum as a dictionary  (no patterns)\n"
    "        =    pattern spectrum as a list        (no patterns)\n"
    "        |    pattern spectrum as three columns (no patterns)\n"
    #ifndef CCN_OVERLAP
    "algo    algorithm variant to use               (default: r)\n"
    #else
    "algo    algorithm variant to use               (default: i)\n"
    #endif
    "        b    basic      basic   version "
                            "(low memory consumption)\n"
    "        f    filter     refined version "
                            "(spike and train filtering)\n"
    "        r    cmrepo     refined version "
                            "(closed/maximal repository)\n"
    #ifndef CCN_OVERLAP
    "        The latter two variants are much faster, "
            "but use more memory.\n"
    #else
    "        i    isect      interval list intersection (repository)\n"
    #endif
    "mode    operation mode indicators/flags        (default: None)\n"
    "        x    do not use perfect extension pruning\n"
    #ifndef CCN_OVERLAP
    "        p    do not sort trains by their length\n"
    #endif
    "        0    use a list/array based closed/maximal repository\n"
    "        1    use a single prefix tree closed/maximal repository\n"
    "             (default: repository is chain of prefix trees)\n"
    #ifdef CCN_OVERLAP
    "mix     mixing parameter for support types     (default: 0)\n"
    #endif
    "border  support border for filtering item sets (default: None)\n"
    "        Must be a list or tuple of minimum support values\n"
    "        per item set size (by which the list/tuple is indexed).\n"
    "returns a list of patterns "
            "(i.e. tuples with one or more elements),\n"
    "        each consisting of a tuple "
            "with a found frequent item set\n"
    "        and the values selected by 'report', "
            "which may be combined\n"
    "        into a tuple or list if report[0] "
            "is '(' or '[', respectively,\n"
    #ifndef CCN_OVERLAP
    "        or a pattern spectrum as a dictionary mapping pairs\n"
    "        (size, supp) to the corresponding occurrence frequency,\n"
    "        as a list of triplets (size, supp, freq) or as three\n"
    "        columns for sizes, support values and frequencies\n"
    #else
    "        or a pattern spectrum as a dictionary "
            "mapping pattern sizes\n"
    "        to the corresponding occurrence support ranges, "
            "as a list\n"
    "        of triplets (size, min. support, max. support) "
            "or as three\n"
    "        columns for sizes and minimum and maximum support values\n"
    #endif
  },
  #ifdef CCN_OVERLAP
  { "cocojim", (PyCFunction)py_cocojim, METH_VARARGS|METH_KEYWORDS,
    "cocojim (trains, target='c', width=0.003, supp=2,\n"
    "         simcov='j', simps=[], sim=0.01, beg=-oo, end=+oo,\n"
    "         zmin=2, zmax=None, report='a', algo='i', mode='',\n"
    "         border=None)\n"
    "COntinuous time ClOsed Neuron Assembly Detection.\n"
    "trains  (spike) train database to mine         (mandatory)\n"
    "        The train database must be an iterable "
            "(list, tuple etc.).\n"
    "        If it is a dictionary, "
            "the keys must be items/event types\n"
    "        that are mapped to iterables of event/spike times.\n"
    "        Otherwise each database element must be a pair, "
            "consisting\n"
    "        of a item/event type and "
            "an iterable of event/spike times.\n"
    "        Each item identifier must be a hashable object.\n"
    "target  type of frequent item sets to find     (default: c)\n"
    "        a/s  all/sets   all     frequent item sets\n"
    "        c    closed     closed  frequent item sets\n"
    "        m    maximal    maximal frequent item sets\n"
    "width   width of time window/influence region  (default: 0.003)\n"
    "supp    minimum support of an item set         (default: 2)\n"
    "simcov  item cover similarity measure          (default: j)\n"
    "        x   none (standard influence map overlap support)\n"
    "        r   Russel & Rao                S_R = s/n\n"
    "        k   Kulczynski                  S_K = s/q\n"
    "        j   Jaccard/Tanimoto            S_J = s/r\n"
    "        d   Dice/Sorensen/Czekanowski   S_D = 2s/(r+s)\n"
    "        s   Sokal & Sneath 1            S_S = s/(r+q)\n"
    "        m   Sokal & Michener/Hamming    S_M = (s+z)/n\n"
    "        f   Faith                       S_F = (s+z/2)/n\n"
    "        t   Rogers & Tanimoto           S_T = (s+z)/(n+q)\n"
    "        n   Sokal & Sneath 2            S_N = 2(s+z)/(n+s+z)\n"
    "        o   Sokal & Sneath 3            S_O = (s+z)/q \n"
    "        b   Baroni-Urbani & Buser       "
                 "S_B = (sqrt(sz)+s)/(sqrt(sz)+r)\n"
    "        g   generic similarity measure  "
                 "S   = (c0s +c1z +c2n +c3sqrt(sz))\n"
    "                                        "
                 "    / (c4s +c5z +c6n +c7sqrt(sz))\n"
    "        all measures are defined in the following quantities:\n"
    "        n   total recording time divided by window width\n"
    "        s   total length of intervals "
                 "with all items in the item set\n"
    "        r   total length of intervals "
                 "with at least one item in the item set\n"
    "        q   total length of intervals "
                 "with some (but not all) items\n"
    "        z   total length of intervals "
                 "with no items in the item set\n"
    "simps   parameters for generic similarity measure\n"
    "        (default: [1,0,0,0,0,-1,1,0])\n"
    "simcov  item cover similarity measure          (default: j)\n"
    "sim     minimum cover similarity measure       (default: 0.01)\n"
    "zmin    minimum number of items per item set   (default: 2)\n"
    "zmax    maximum number of items per item set   "
            "(default: no limit)\n"
    "report  values to report with an item set      (default: a)\n"
    "        if mode does not contain the letter 'e':\n"
    "          a   (absolute) item set support\n"
    "          w   item cover similarity as a fraction\n"
    "          W   item cover similarity as a percentage\n"
    "        if mode contains the letter 'e':\n"
    "          w   (absolute) item set support\n"
    "          a   item cover similarity as a fraction\n"
    "          s   item cover similarity as a fraction\n"
    "          S   item cover similarity as a percentage\n"
    "        in any case:\n"
    "        (    combine values in a tuple (must be first character)\n"
    "        [    combine values in a list  (must be first character)\n"
    "        #    pattern spectrum as a dictionary  (no patterns)\n"
    "        =    pattern spectrum as a list        (no patterns)\n"
    "        |    pattern spectrum as three columns (no patterns)\n"
    "algo    algorithm variant to use               (default: i)\n"
    "        i    isect      interval list intersection (repository)\n"
    "mode    operation mode indicators/flags        (default: None)\n"
    "        x    do not use perfect extension pruning\n"
    "        0    use a list/array based closed/maximal repository\n"
    "        e    exchange roles of support and cover similarity\n"
    "             (only for reporting and "
                  "pattern spectrum generation)\n"
    "mode    special flags for search/reporting     (default: None)\n"
    "border  support border for filtering item sets (default: None)\n"
    "        Must be a list or tuple of minimum support values\n"
    "        per item set size (by which the list/tuple is indexed).\n"
    "returns a list of patterns "
            "(i.e. tuples with one or more elements),\n"
    "        each consisting of a tuple "
            "with a found frequent item set\n"
    "        and the values selected by 'report', "
            "which may be combined\n"
    "        into a tuple or list "
            "if report[0] is '(' or '[', respectively,\n"
    "        or a pattern spectrum as a dictionary "
            "mapping pattern sizes\n"
    "        to the corresponding occurrence support ranges, "
            "as a list\n"
    "        of triplets (size, min. support, max. support) "
            "or as three\n"
    "        columns for sizes and minimum and maximum support values\n"
  },
  #endif  /* #ifdef CCN_OVERLAP */
  { "genpsp", (PyCFunction)py_genpsp, METH_VARARGS|METH_KEYWORDS,
    "genpsp (trains, target='c', width=0.003, supp=2, "
            "zmin=2, zmax=None,\n"
    #ifdef CCN_OVERLAP
    "        report='#', algo='i', mode='', cnt=1000, "
            "beg=-oo, end=oo,\n"
    #else
    "        report='#', algo='r', mode='', cnt=1000, "
            "beg=-oo, end=oo,\n"
    #endif
    "        surr='p', rand='u', sigma=0.005, delta=0.03, "
            "seed=0, cpus=0)\n"
    "Generate a pattern spectrum from surrogate data sets.\n"
    "trains  (spike) train database to mine         (mandatory)\n"
    "        The train database must be an iterable "
            "(list, tuple etc.).\n"
    "        If it is a dictionary, "
            "the keys must be items/event types\n"
    "        that are mapped to iterables of event/spike times.\n"
    "        Otherwise each database element must be a pair, "
            "consisting\n"
    "        of a item/event type "
            "and an iterable of event/spike times.\n"
    "        Each item identifier must be a hashable object.\n"
    "target  type of frequent item sets to find     (default: c)\n"
    "        a/s  all/sets   all     frequent item sets\n"
    "        c    closed     closed  frequent item sets\n"
    "        m    maximal    maximal frequent item sets\n"
    #ifndef CCN_OVERLAP         /* if standard (MIS) support */
    "        ap   perms      all     frequent item permutations\n"
    "        cp   clsperms   closed  frequent item permutations\n"
    "        mp   maxperms   maximal frequent item permutations\n"
    "        as   seqs       all     frequent item sequences\n"
    "        cs   clsseqs    closed  frequent item sequences\n"
    "        ms   maxseqs    maximal frequent item sequences\n"
    #endif
    "width   width of time window/influence region  (default: 0.003)\n"
    "supp    minimum support of an item set         (default: 2)\n"
    "zmin    minimum number of items per item set   (default: 2)\n"
    "zmax    maximum number of items per item set   "
            "(default: no limit)\n"
    "report  pattern spectrum reporting format      (default: #)\n"
    "        #    pattern spectrum as a dictionary "
                 "(size,supp) -> freq\n"
    "        =    pattern spectrum as a list of triplets\n"
    "        |    pattern spectrum as three columns\n"
    #ifndef CCN_OVERLAP
    "algo    algorithm variant to use               (default: r)\n"
    #else
    "algo    algorithm variant to use               (default: i)\n"
    #endif
    "        b    basic      basic   version "
                            "(low memory consumption)\n"
    "        f    filter     refined version "
                            "(spike and train filtering)\n"
    "        r    cmrepo     refined version "
                            "(closed/maximal repository)\n"
    #ifndef CCN_OVERLAP
    "        The latter two variants are much faster, "
            "but use more memory.\n"
    #else
    "        i    isect      interval list intersection (repository)\n"
    #endif
    "mode    operation mode indicators/flags        (default: None)\n"
    "        x    do not use perfect extension pruning\n"
    #ifndef CCN_OVERLAP
    "        p    do not sort trains by their length\n"
    #else
    "mix     mixing parameter for support types     (default: 0)\n"
    #endif
    "cnt     number of surrogate data sets          (default: 1000)\n"
    "beg     beginning of range of events           (default: -oo)\n"
    "end     end       of range of events           (default: +oo)\n"
    "surr    surrogate data generation method       (default: p)\n"
    "        i    ident      identity (keep original data)\n"
    "        r    random     event/spike time randomization\n"
    "        d    dither     event/spike dithering/displacement\n"
    "        s    shift      train shifting/dithering\n"
    "        k    kernest    sampling from a kernel estimate\n"
    "        p    permute    dithered event/spike permutation\n"
    "        b    block      dithered blocked "
                            "event/spike permutations\n"
    "rand    random function density identifier     (default: u)\n"
    "        u    uniform    uniform density (= rectangular)\n"
    "        r    rect       rectangular density (= uniform)\n"
    "        t    triang     symmetric triangular density\n"
    "        g    gauss      Gaussian density (= normal)\n"
    "        n    normal     normal density (= Gaussian)\n"
    "sigma   random dispersion parameter            (default: 0.005)\n"
    "        (standard deviation or half the base width of density)\n"
    "delta   block size for block permutations      (default: 0.03)\n"
    "seed    seed for random number generator       (default: 0)\n"
    "        (seed = 0: use system time as a seed)\n"
    "cpus    number of cpus to use                  (default: 0)\n"
    "        (cpus = 0: determine number of cores automatically)\n"
    #ifndef CCN_OVERLAP
    "returns a pattern spectrum as a dictionary mapping pairs\n"
    "        (size, supp) to the corresponding occurrence frequency,\n"
    "        as a list of triplets (size, supp, freq) or as three\n"
    "        columns for sizes, support values and frequencies\n"
    #else
    "returns a pattern spectrum as a dictionary mapping pattern sizes\n"
    "        to the corresponding occurrence support ranges, "
            "as a list\n"
    "        of triplets (size, min. support, max. support) "
            "or as three\n"
    "        columns for sizes and minimum and maximum support values\n"
    #endif
  },
#ifndef CCN_OVERLAP
  { "estpsp", (PyCFunction)py_estpsp, METH_VARARGS|METH_KEYWORDS,
    "estpsp (trains, target='a', width=0.003, supp=2, "
            "zmin=2, zmax=None,\n"
    "        report='#', equiv=10000, alpha=0.5, smpls=1000, seed=0)\n"
    "Estimate a pattern spectrum from data characteristics.\n"
    "trains  (spike) train database to mine         (mandatory)\n"
    "        The train database must be an iterable "
            "(list, tuple etc.).\n"
    "        If it is a dictionary, "
            "the keys must be items/event types\n"
    "        that are mapped to iterables of event/spike times.\n"
    "        Otherwise each database element must be a pair, "
            "consisting\n"
    "        of a item/event type "
            "and an iterable of event/spike times.\n"
    "        Each item identifier must be a hashable object.\n"
    "target  type of frequent item sets to find     (default: a)\n"
    "        a/s  all/sets   frequent item sets\n"
    "        ap   perms      frequent item permutations\n"
    "        as   seqs       frequent item sequences\n"
    "width   width of time window/influence region  (default: 0.003)\n"
    "supp    minimum support of an item set         (default: 2)\n"
    "zmin    minimum number of items per item set   (default: 2)\n"
    "zmax    maximum number of items per item set   "
            "(default: no limit)\n"
    "report  pattern spectrum reporting format      (default: #)\n"
    "        #    pattern spectrum as a dictionary "
                 "(size,supp) -> freq\n"
    "        =    pattern spectrum as a list of triplets\n"
    "        |    pattern spectrum as three columns\n"
    "equiv   equivalent number of surrogates        (default: 10000)\n"
    "alpha   probability dispersion factor          (default: 0.5)\n"
    "smpls   number of samples per item set size    (default: 1000)\n"
    "seed    seed for random number generator       (default: 0)\n"
    "        (seed = 0: use system time as a seed)\n"
    "returns a pattern spectrum as a dictionary mapping pairs\n"
    "        (size, supp) to the corresponding occurrence frequency,\n"
    "        as a list of triplets (size, supp, freq) or as three\n"
    "        columns for sizes, support values and frequencies\n"
  },
#endif
  { "psp2bdr", (PyCFunction)py_psp2bdr, METH_VARARGS|METH_KEYWORDS,
    "psp2bdr (psp)\n"
    "Extract a decision border from a pattern spectrum.\n"
    "psp     a pattern spectrum as returned by one of the functions\n"
    "        genpsp() or estpsp() in the form of dictionary or a list\n"
    "        of triplets (the three column format is not supported;\n"
    "        for a 3-column pattern spectrum, use psp2bdr(zip(*psp)))\n"
    "returns a decision border as a list of minimum support values\n"
    "        that is to be index by the pattern size (starting at 0)\n"
  },
  { "patred", (PyCFunction)py_patred, METH_VARARGS|METH_KEYWORDS,
    "patred (pats, method='S', border=None, addis=True)\n"
    "Perform pattern set reduction based on a preference relation.\n"
    "pats    set of patterns to reduced             (mandatory)\n"
    "        The pattern set must be an iterable (list, tuple etc.).\n"
    "        If it is a dictionary, the keys must be item sets\n"
    "        that are mapped to their support (number).\n"
    "        Otherwise each iterable element must be a pair, "
            "consisting\n"
    "        of a item set as an iterable "
            "of item and a support (number).\n"
    "        Each item identifier must be a hashable object.\n"
    "method  index or identifier of pattern comparison method or\n"
    "        a pattern comparison function "
            "patcmp(zA, cA, zB, cB, border)\n"
    "border  detection border as a list of "
            "minimum support values per size\n"
    "addis   whether to add intersections of patterns\n"
    "returns a reduced set of patterns, "
            "either as a dictionary that maps\n"
    "        item sets to support values "
            "(if the input was a dictionary)\n"
    "        or as a list of patterns in the form in which they were\n"
    "        passed into this function (all parts are maintained)"
  },
  { NULL }                      /* sentinel */
};

/*----------------------------------------------------------------------
  Initialization Function
----------------------------------------------------------------------*/
#define COCO_DESC \
  "COntinuous time ClOsed Neuron Assembly Detection for Python\n" \
  "version 2.44 (2016.11.21)      (c) 2013-2016   Christian Borgelt"

#if PY_MAJOR_VERSION >= 3

#ifndef CCNOVERLAP
static struct PyModuleDef cocodef = {
  PyModuleDef_HEAD_INIT, "coco",   COCO_DESC,
  -1, coco_methods, NULL, NULL, NULL, NULL
};
#else
static struct PyModuleDef cocodef = {
  PyModuleDef_HEAD_INIT, "ccnovl", COCO_DESC,
  -1, coco_methods, NULL, NULL, NULL, NULL
};
#endif

#ifndef CCN_OVERLAP
PyMODINIT_FUNC PyInit_coco (void)
{ return PyModule_Create(&cocodef); }
#else
PyMODINIT_FUNC PyInit_ccnovl (void)
{ return PyModule_Create(&cocodef); }
#endif

#else

#ifndef CCN_OVERLAP
PyMODINIT_FUNC initcoco (void)
{ Py_InitModule3("coco",   coco_methods, COCO_DESC); }
#else
PyMODINIT_FUNC initccnovl (void)
{ Py_InitModule3("ccnovl", coco_methods, COCO_DESC); }
#endif

#endif
